# Copyright (c) OpenMMLab. All rights reserved.
import os
import re
from typing import Optional

import imageio
import numpy as np
import torch
import torchvision
from diffusers.pipelines.paint_by_example import PaintByExampleImageEncoder
from diffusers.pipelines.stable_diffusion import StableUnCLIPImageNormalizer
from diffusers.schedulers import DDPMScheduler
from einops import rearrange
from transformers import (CLIPImageProcessor, CLIPTextModel, CLIPVisionConfig,
                          CLIPVisionModelWithProjection)


def shave_segments(path, n_shave_prefix_segments=1):
    """Removes segments.

    Positive values shave the first segments, negative shave the last segments.
    """
    if n_shave_prefix_segments >= 0:
        return '.'.join(path.split('.')[n_shave_prefix_segments:])
    else:
        return '.'.join(path.split('.')[:n_shave_prefix_segments])


def renew_resnet_paths(old_list, n_shave_prefix_segments=0):
    """Updates paths inside resnets to the new naming scheme (local
    renaming)"""
    mapping = []
    for old_item in old_list:
        new_item = old_item.replace('in_layers.0', 'norm1')
        new_item = new_item.replace('in_layers.2', 'conv1')

        new_item = new_item.replace('out_layers.0', 'norm2')
        new_item = new_item.replace('out_layers.3', 'conv2')

        new_item = new_item.replace('emb_layers.1', 'time_emb_proj')
        new_item = new_item.replace('skip_connection', 'conv_shortcut')

        new_item = shave_segments(
            new_item, n_shave_prefix_segments=n_shave_prefix_segments)

        mapping.append({'old': old_item, 'new': new_item})

    return mapping


def renew_vae_resnet_paths(old_list, n_shave_prefix_segments=0):
    """Updates paths inside resnets to the new naming scheme (local
    renaming)"""
    mapping = []
    for old_item in old_list:
        new_item = old_item

        new_item = new_item.replace('nin_shortcut', 'conv_shortcut')
        new_item = shave_segments(
            new_item, n_shave_prefix_segments=n_shave_prefix_segments)

        mapping.append({'old': old_item, 'new': new_item})

    return mapping


def renew_attention_paths(old_list, n_shave_prefix_segments=0):
    """Updates paths inside attentions to the new naming scheme (local
    renaming)"""
    mapping = []
    for old_item in old_list:
        new_item = old_item

        mapping.append({'old': old_item, 'new': new_item})

    return mapping


def renew_vae_attention_paths(old_list, n_shave_prefix_segments=0):
    """Updates paths inside attentions to the new naming scheme (local
    renaming)"""
    mapping = []
    for old_item in old_list:
        new_item = old_item

        new_item = new_item.replace('norm.weight', 'group_norm.weight')
        new_item = new_item.replace('norm.bias', 'group_norm.bias')

        new_item = new_item.replace('q.weight', 'query.weight')
        new_item = new_item.replace('q.bias', 'query.bias')

        new_item = new_item.replace('k.weight', 'key.weight')
        new_item = new_item.replace('k.bias', 'key.bias')

        new_item = new_item.replace('v.weight', 'value.weight')
        new_item = new_item.replace('v.bias', 'value.bias')

        new_item = new_item.replace('proj_out.weight', 'proj_attn.weight')
        new_item = new_item.replace('proj_out.bias', 'proj_attn.bias')

        new_item = shave_segments(
            new_item, n_shave_prefix_segments=n_shave_prefix_segments)

        mapping.append({'old': old_item, 'new': new_item})

    return mapping


def assign_to_checkpoint(paths,
                         checkpoint,
                         old_checkpoint,
                         attention_paths_to_split=None,
                         additional_replacements=None,
                         config=None):
    """This does the final conversion step: take locally converted weights and
    apply a global renaming to them. It splits attention layers, and takes into
    account additional replacements that may arise.

    Assigns the weights to the new checkpoint.
    """
    assert isinstance(
        paths, list
    ), "Paths should be a list of dicts containing 'old' and 'new' keys."

    # Splits the attention layers into three variables.
    if attention_paths_to_split is not None:
        for path, path_map in attention_paths_to_split.items():
            old_tensor = old_checkpoint[path]
            channels = old_tensor.shape[0] // 3

            target_shape = (-1,
                            channels) if len(old_tensor.shape) == 3 else (-1)

            num_heads = old_tensor.shape[0] // config['num_head_channels'] // 3

            old_tensor = old_tensor.reshape((num_heads, 3 * channels //
                                             num_heads) + old_tensor.shape[1:])
            query, key, value = old_tensor.split(channels // num_heads, dim=1)

            checkpoint[path_map['query']] = query.reshape(target_shape)
            checkpoint[path_map['key']] = key.reshape(target_shape)
            checkpoint[path_map['value']] = value.reshape(target_shape)

    for path in paths:
        new_path = path['new']

        # These have already been assigned
        if attention_paths_to_split is not None and (
                new_path in attention_paths_to_split):
            continue
        # Global renaming happens here
        new_path = new_path.replace('middle_block.0', 'mid_block.resnets.0')
        new_path = new_path.replace('middle_block.1', 'mid_block.attentions.0')
        new_path = new_path.replace('middle_block.2', 'mid_block.resnets.1')

        if additional_replacements is not None:
            for replacement in additional_replacements:
                new_path = new_path.replace(replacement['old'],
                                            replacement['new'])

        # proj_attn.weight has to be converted from conv 1D to linear
        if 'proj_attn.weight' in new_path:
            checkpoint[new_path] = old_checkpoint[path['old']][:, :, 0]
        else:
            checkpoint[new_path] = old_checkpoint[path['old']]


def conv_attn_to_linear(checkpoint):
    keys = list(checkpoint.keys())
    attn_keys = ['query.weight', 'key.weight', 'value.weight']
    for key in keys:
        if '.'.join(key.split('.')[-2:]) in attn_keys:
            if checkpoint[key].ndim > 2:
                checkpoint[key] = checkpoint[key][:, :, 0, 0]
        elif 'proj_attn.weight' in key:
            if checkpoint[key].ndim > 2:
                checkpoint[key] = checkpoint[key][:, :, 0]


def create_unet_diffusers_config(original_config,
                                 image_size: int,
                                 controlnet=False):
    """Creates a config for the diffusers based on the config of the LDM
    model."""
    if controlnet:
        unet_params = original_config.model.params.control_stage_config.params
    else:
        unet_params = original_config.model.params.unet_config.params

    vae_params = \
        original_config.model.params.first_stage_config.params.ddconfig

    block_out_channels = [
        unet_params.model_channels * mult for mult in unet_params.channel_mult
    ]

    down_block_types = []
    resolution = 1
    for i in range(len(block_out_channels)):
        block_type = 'CrossAttnDownBlock2D' \
            if resolution in unet_params.attention_resolutions \
            else 'DownBlock2D'
        down_block_types.append(block_type)
        if i != len(block_out_channels) - 1:
            resolution *= 2

    up_block_types = []
    for i in range(len(block_out_channels)):
        block_type = 'CrossAttnUpBlock2D' \
            if resolution in unet_params.attention_resolutions \
            else 'UpBlock2D'
        up_block_types.append(block_type)
        resolution //= 2

    vae_scale_factor = 2**(len(vae_params.ch_mult) - 1)

    head_dim = unet_params.num_heads if 'num_heads' in unet_params else None
    use_linear_projection = (
        unet_params.use_linear_in_transformer
        if 'use_linear_in_transformer' in unet_params else False)
    if use_linear_projection:
        # stable diffusion 2-base-512 and 2-768
        if head_dim is None:
            head_dim = [5, 10, 20, 20]

    class_embed_type = None
    projection_class_embeddings_input_dim = None

    if 'num_classes' in unet_params:
        if unet_params.num_classes == 'sequential':
            class_embed_type = 'projection'
            assert 'adm_in_channels' in unet_params
            projection_class_embeddings_input_dim = unet_params.adm_in_channels
        else:
            raise NotImplementedError(
                f'Unknown conditional unet num_classes config: '
                f'{unet_params.num_classes}')

    config = {
        'sample_size':
        image_size // vae_scale_factor,
        'in_channels':
        unet_params.in_channels,
        'down_block_types':
        tuple(down_block_types),
        'block_out_channels':
        tuple(block_out_channels),
        'layers_per_block':
        unet_params.num_res_blocks,
        'cross_attention_dim':
        unet_params.context_dim,
        'attention_head_dim':
        head_dim,
        'use_linear_projection':
        use_linear_projection,
        'class_embed_type':
        class_embed_type,
        'projection_class_embeddings_input_dim':
        projection_class_embeddings_input_dim,
    }

    if not controlnet:
        config['out_channels'] = unet_params.out_channels
        config['up_block_types'] = tuple(up_block_types)

    return config


def create_vae_diffusers_config(original_config, image_size: int):
    """Creates a config for the diffusers based on the config of the LDM
    model."""
    vae_params = \
        original_config.model.params.first_stage_config.params.ddconfig
    _ = original_config.model.params.first_stage_config.params.embed_dim

    block_out_channels = [vae_params.ch * mult for mult in vae_params.ch_mult]
    down_block_types = ['DownEncoderBlock2D'] * len(block_out_channels)
    up_block_types = ['UpDecoderBlock2D'] * len(block_out_channels)

    config = {
        'sample_size': image_size,
        'in_channels': vae_params.in_channels,
        'out_channels': vae_params.out_ch,
        'down_block_types': tuple(down_block_types),
        'up_block_types': tuple(up_block_types),
        'block_out_channels': tuple(block_out_channels),
        'latent_channels': vae_params.z_channels,
        'layers_per_block': vae_params.num_res_blocks,
    }
    return config


def convert_ldm_unet_checkpoint(checkpoint,
                                config,
                                path=None,
                                extract_ema=False,
                                controlnet=False):
    """Takes a state dict and a config, and returns a converted checkpoint."""

    # extract state_dict for UNet
    unet_state_dict = {}
    keys = list(checkpoint.keys())

    if controlnet:
        unet_key = 'control_model.'
    else:
        unet_key = 'model.diffusion_model.'

    # at least a 100 parameters have to start with `model_ema`
    # in order for the checkpoint to be EMA
    if sum(k.startswith('model_ema') for k in keys) > 100 and extract_ema:
        print(f'Checkpoint {path} has both EMA and non-EMA weights.')
        print('In this conversion only the EMA weights are extracted. '
              'If you want to instead extract the non-EMA'
              ' weights (useful to continue fine-tuning), please make sure to '
              'remove the `--extract_ema` flag.')
        for key in keys:
            if key.startswith('model.diffusion_model'):
                flat_ema_key = 'model_ema.' + ''.join(key.split('.')[1:])
                unet_state_dict[key.replace(unet_key,
                                            '')] = checkpoint.pop(flat_ema_key)
    else:
        if sum(k.startswith('model_ema') for k in keys) > 100:
            print('In this conversion only the non-EMA weights '
                  'are extracted. If you want to instead extract the EMA'
                  ' weights (usually better for inference), please'
                  ' make sure to add the `--extract_ema` flag.')

        for key in keys:
            if key.startswith(unet_key):
                unet_state_dict[key.replace(unet_key,
                                            '')] = checkpoint.pop(key)

    new_checkpoint = {}

    new_checkpoint['time_embedding.linear_1.weight'] = unet_state_dict[
        'time_embed.0.weight']
    new_checkpoint['time_embedding.linear_1.bias'] = unet_state_dict[
        'time_embed.0.bias']
    new_checkpoint['time_embedding.linear_2.weight'] = unet_state_dict[
        'time_embed.2.weight']
    new_checkpoint['time_embedding.linear_2.bias'] = unet_state_dict[
        'time_embed.2.bias']

    if config['class_embed_type'] is None:
        # No parameters to port
        ...
    elif config['class_embed_type'] == 'timestep' or config[
            'class_embed_type'] == 'projection':
        new_checkpoint['class_embedding.linear_1.weight'] = unet_state_dict[
            'label_emb.0.0.weight']
        new_checkpoint['class_embedding.linear_1.bias'] = unet_state_dict[
            'label_emb.0.0.bias']
        new_checkpoint['class_embedding.linear_2.weight'] = unet_state_dict[
            'label_emb.0.2.weight']
        new_checkpoint['class_embedding.linear_2.bias'] = unet_state_dict[
            'label_emb.0.2.bias']
    else:
        raise NotImplementedError(
            f"Not implemented `class_embed_type`: {config['class_embed_type']}"
        )

    new_checkpoint['conv_in.weight'] = unet_state_dict[
        'input_blocks.0.0.weight']
    new_checkpoint['conv_in.bias'] = unet_state_dict['input_blocks.0.0.bias']

    if not controlnet:
        new_checkpoint['conv_norm_out.weight'] = unet_state_dict[
            'out.0.weight']
        new_checkpoint['conv_norm_out.bias'] = unet_state_dict['out.0.bias']
        new_checkpoint['conv_out.weight'] = unet_state_dict['out.2.weight']
        new_checkpoint['conv_out.bias'] = unet_state_dict['out.2.bias']

    # Retrieves the keys for the input blocks only
    num_input_blocks = len({
        '.'.join(layer.split('.')[:2])
        for layer in unet_state_dict if 'input_blocks' in layer
    })
    input_blocks = {
        layer_id:
        [key for key in unet_state_dict if f'input_blocks.{layer_id}' in key]
        for layer_id in range(num_input_blocks)
    }

    # Retrieves the keys for the middle blocks only
    num_middle_blocks = len({
        '.'.join(layer.split('.')[:2])
        for layer in unet_state_dict if 'middle_block' in layer
    })
    middle_blocks = {
        layer_id:
        [key for key in unet_state_dict if f'middle_block.{layer_id}' in key]
        for layer_id in range(num_middle_blocks)
    }

    # Retrieves the keys for the output blocks only
    num_output_blocks = len({
        '.'.join(layer.split('.')[:2])
        for layer in unet_state_dict if 'output_blocks' in layer
    })
    output_blocks = {
        layer_id:
        [key for key in unet_state_dict if f'output_blocks.{layer_id}' in key]
        for layer_id in range(num_output_blocks)
    }

    for i in range(1, num_input_blocks):
        block_id = (i - 1) // (config['layers_per_block'] + 1)
        layer_in_block_id = (i - 1) % (config['layers_per_block'] + 1)

        resnets = [
            key for key in input_blocks[i] if f'input_blocks.{i}.0' in key
            and f'input_blocks.{i}.0.op' not in key
        ]
        attentions = [
            key for key in input_blocks[i] if f'input_blocks.{i}.1' in key
        ]

        if f'input_blocks.{i}.0.op.weight' in unet_state_dict:
            new_checkpoint[
                f'down_blocks.{block_id}.downsamplers.0.conv.weight'] = \
                unet_state_dict.pop(f'input_blocks.{i}.0.op.weight')
            new_checkpoint[
                f'down_blocks.{block_id}.downsamplers.0.conv.bias'] = \
                unet_state_dict.pop(f'input_blocks.{i}.0.op.bias')

        paths = renew_resnet_paths(resnets)
        meta_path = {
            'old': f'input_blocks.{i}.0',
            'new': f'down_blocks.{block_id}.resnets.{layer_in_block_id}'
        }
        assign_to_checkpoint(
            paths,
            new_checkpoint,
            unet_state_dict,
            additional_replacements=[meta_path],
            config=config)

        if len(attentions):
            paths = renew_attention_paths(attentions)
            meta_path = {
                'old': f'input_blocks.{i}.1',
                'new': f'down_blocks.{block_id}.attentions.{layer_in_block_id}'
            }
            assign_to_checkpoint(
                paths,
                new_checkpoint,
                unet_state_dict,
                additional_replacements=[meta_path],
                config=config)

    resnet_0 = middle_blocks[0]
    attentions = middle_blocks[1]
    resnet_1 = middle_blocks[2]

    resnet_0_paths = renew_resnet_paths(resnet_0)
    assign_to_checkpoint(
        resnet_0_paths, new_checkpoint, unet_state_dict, config=config)

    resnet_1_paths = renew_resnet_paths(resnet_1)
    assign_to_checkpoint(
        resnet_1_paths, new_checkpoint, unet_state_dict, config=config)

    attentions_paths = renew_attention_paths(attentions)
    meta_path = {'old': 'middle_block.1', 'new': 'mid_block.attentions.0'}
    assign_to_checkpoint(
        attentions_paths,
        new_checkpoint,
        unet_state_dict,
        additional_replacements=[meta_path],
        config=config)

    for i in range(num_output_blocks):
        block_id = i // (config['layers_per_block'] + 1)
        layer_in_block_id = i % (config['layers_per_block'] + 1)
        output_block_layers = [
            shave_segments(name, 2) for name in output_blocks[i]
        ]
        output_block_list = {}

        for layer in output_block_layers:
            layer_id, layer_name = layer.split('.')[0], shave_segments(
                layer, 1)
            if layer_id in output_block_list:
                output_block_list[layer_id].append(layer_name)
            else:
                output_block_list[layer_id] = [layer_name]

        if len(output_block_list) > 1:
            resnets = [
                key for key in output_blocks[i]
                if f'output_blocks.{i}.0' in key
            ]
            attentions = [
                key for key in output_blocks[i]
                if f'output_blocks.{i}.1' in key
            ]

            resnet_0_paths = renew_resnet_paths(resnets)
            paths = renew_resnet_paths(resnets)

            meta_path = {
                'old': f'output_blocks.{i}.0',
                'new': f'up_blocks.{block_id}.resnets.{layer_in_block_id}'
            }
            assign_to_checkpoint(
                paths,
                new_checkpoint,
                unet_state_dict,
                additional_replacements=[meta_path],
                config=config)

            output_block_list = {
                k: sorted(v)
                for k, v in output_block_list.items()
            }
            if ['conv.bias', 'conv.weight'] in output_block_list.values():
                index = list(output_block_list.values()).index(
                    ['conv.bias', 'conv.weight'])
                new_checkpoint[
                    f'up_blocks.{block_id}.upsamplers.0.conv.weight'] = \
                    unet_state_dict[
                        f'output_blocks.{i}.{index}.conv.weight']
                new_checkpoint[
                    f'up_blocks.{block_id}.upsamplers.0.conv.bias'] = \
                    unet_state_dict[
                        f'output_blocks.{i}.{index}.conv.bias']

                # Clear attentions as they have been attributed above.
                if len(attentions) == 2:
                    attentions = []

            if len(attentions):
                paths = renew_attention_paths(attentions)
                meta_path = {
                    'old': f'output_blocks.{i}.1',
                    'new':
                    f'up_blocks.{block_id}.attentions.{layer_in_block_id}',
                }
                assign_to_checkpoint(
                    paths,
                    new_checkpoint,
                    unet_state_dict,
                    additional_replacements=[meta_path],
                    config=config)
        else:
            resnet_0_paths = renew_resnet_paths(
                output_block_layers, n_shave_prefix_segments=1)
            for path in resnet_0_paths:
                old_path = '.'.join(['output_blocks', str(i), path['old']])
                new_path = '.'.join([
                    'up_blocks',
                    str(block_id), 'resnets',
                    str(layer_in_block_id), path['new']
                ])

                new_checkpoint[new_path] = unet_state_dict[old_path]

    if controlnet:
        # conditioning embedding

        orig_index = 0

        new_checkpoint[
            'controlnet_cond_embedding.conv_in.weight'] = unet_state_dict.pop(
                f'input_hint_block.{orig_index}.weight')
        new_checkpoint[
            'controlnet_cond_embedding.conv_in.bias'] = unet_state_dict.pop(
                f'input_hint_block.{orig_index}.bias')

        orig_index += 2

        diffusers_index = 0

        while diffusers_index < 6:
            new_checkpoint[f'controlnet_cond_embedding.blocks'
                           f'.{diffusers_index}.weight'] = (
                               unet_state_dict.pop(
                                   f'input_hint_block.{orig_index}.weight'))
            new_checkpoint[
                f'controlnet_cond_embedding.blocks.{diffusers_index}.bias'] = (
                    unet_state_dict.pop(f'input_hint_block.{orig_index}.bias'))
            diffusers_index += 1
            orig_index += 2

        new_checkpoint[
            'controlnet_cond_embedding.conv_out.weight'] = unet_state_dict.pop(
                f'input_hint_block.{orig_index}.weight')
        new_checkpoint[
            'controlnet_cond_embedding.conv_out.bias'] = unet_state_dict.pop(
                f'input_hint_block.{orig_index}.bias')

        # down blocks
        for i in range(num_input_blocks):
            new_checkpoint[
                f'controlnet_down_blocks.{i}.weight'] = unet_state_dict.pop(
                    f'zero_convs.{i}.0.weight')
            new_checkpoint[
                f'controlnet_down_blocks.{i}.bias'] = unet_state_dict.pop(
                    f'zero_convs.{i}.0.bias')

        # mid block
        new_checkpoint['controlnet_mid_block.weight'] = unet_state_dict.pop(
            'middle_block_out.0.weight')
        new_checkpoint['controlnet_mid_block.bias'] = unet_state_dict.pop(
            'middle_block_out.0.bias')

    return new_checkpoint


def convert_ldm_vae_checkpoint(checkpoint, config):
    # extract state dict for VAE
    vae_state_dict = {}
    vae_key = 'first_stage_model.'
    keys = list(checkpoint.keys())
    for key in keys:
        if key.startswith(vae_key):
            vae_state_dict[key.replace(vae_key, '')] = checkpoint.get(key)

    new_checkpoint = {}

    new_checkpoint['encoder.conv_in.weight'] = vae_state_dict[
        'encoder.conv_in.weight']
    new_checkpoint['encoder.conv_in.bias'] = vae_state_dict[
        'encoder.conv_in.bias']
    new_checkpoint['encoder.conv_out.weight'] = vae_state_dict[
        'encoder.conv_out.weight']
    new_checkpoint['encoder.conv_out.bias'] = vae_state_dict[
        'encoder.conv_out.bias']
    new_checkpoint['encoder.conv_norm_out.weight'] = vae_state_dict[
        'encoder.norm_out.weight']
    new_checkpoint['encoder.conv_norm_out.bias'] = vae_state_dict[
        'encoder.norm_out.bias']

    new_checkpoint['decoder.conv_in.weight'] = vae_state_dict[
        'decoder.conv_in.weight']
    new_checkpoint['decoder.conv_in.bias'] = vae_state_dict[
        'decoder.conv_in.bias']
    new_checkpoint['decoder.conv_out.weight'] = vae_state_dict[
        'decoder.conv_out.weight']
    new_checkpoint['decoder.conv_out.bias'] = vae_state_dict[
        'decoder.conv_out.bias']
    new_checkpoint['decoder.conv_norm_out.weight'] = vae_state_dict[
        'decoder.norm_out.weight']
    new_checkpoint['decoder.conv_norm_out.bias'] = vae_state_dict[
        'decoder.norm_out.bias']

    new_checkpoint['quant_conv.weight'] = vae_state_dict['quant_conv.weight']
    new_checkpoint['quant_conv.bias'] = vae_state_dict['quant_conv.bias']
    new_checkpoint['post_quant_conv.weight'] = vae_state_dict[
        'post_quant_conv.weight']
    new_checkpoint['post_quant_conv.bias'] = vae_state_dict[
        'post_quant_conv.bias']

    # Retrieves the keys for the encoder down blocks only
    num_down_blocks = len({
        '.'.join(layer.split('.')[:3])
        for layer in vae_state_dict if 'encoder.down' in layer
    })
    down_blocks = {
        layer_id: [key for key in vae_state_dict if f'down.{layer_id}' in key]
        for layer_id in range(num_down_blocks)
    }

    # Retrieves the keys for the decoder up blocks only
    num_up_blocks = len({
        '.'.join(layer.split('.')[:3])
        for layer in vae_state_dict if 'decoder.up' in layer
    })
    up_blocks = {
        layer_id: [key for key in vae_state_dict if f'up.{layer_id}' in key]
        for layer_id in range(num_up_blocks)
    }

    for i in range(num_down_blocks):
        resnets = [
            key for key in down_blocks[i]
            if f'down.{i}' in key and f'down.{i}.downsample' not in key
        ]

        if f'encoder.down.{i}.downsample.conv.weight' in vae_state_dict:
            new_checkpoint[
                f'encoder.down_blocks.{i}.downsamplers.0.conv.weight'] = \
                vae_state_dict.pop(f'encoder.down.{i}.downsample.conv.weight')
            new_checkpoint[
                f'encoder.down_blocks.{i}.downsamplers.0.conv.bias'] = \
                vae_state_dict.pop(f'encoder.down.{i}.downsample.conv.bias')

        paths = renew_vae_resnet_paths(resnets)
        meta_path = {
            'old': f'down.{i}.block',
            'new': f'down_blocks.{i}.resnets'
        }
        assign_to_checkpoint(
            paths,
            new_checkpoint,
            vae_state_dict,
            additional_replacements=[meta_path],
            config=config)

    mid_resnets = [key for key in vae_state_dict if 'encoder.mid.block' in key]
    num_mid_res_blocks = 2
    for i in range(1, num_mid_res_blocks + 1):
        resnets = [
            key for key in mid_resnets if f'encoder.mid.block_{i}' in key
        ]

        paths = renew_vae_resnet_paths(resnets)
        meta_path = {
            'old': f'mid.block_{i}',
            'new': f'mid_block.resnets.{i - 1}'
        }
        assign_to_checkpoint(
            paths,
            new_checkpoint,
            vae_state_dict,
            additional_replacements=[meta_path],
            config=config)

    mid_attentions = [
        key for key in vae_state_dict if 'encoder.mid.attn' in key
    ]
    paths = renew_vae_attention_paths(mid_attentions)
    meta_path = {'old': 'mid.attn_1', 'new': 'mid_block.attentions.0'}
    assign_to_checkpoint(
        paths,
        new_checkpoint,
        vae_state_dict,
        additional_replacements=[meta_path],
        config=config)
    conv_attn_to_linear(new_checkpoint)

    for i in range(num_up_blocks):
        block_id = num_up_blocks - 1 - i
        resnets = [
            key for key in up_blocks[block_id]
            if f'up.{block_id}' in key and f'up.{block_id}.upsample' not in key
        ]

        if f'decoder.up.{block_id}.upsample.conv.weight' in vae_state_dict:
            new_checkpoint[
                f'decoder.up_blocks.{i}.upsamplers.0.conv.weight'] = \
                vae_state_dict[f'decoder.up.{block_id}.upsample.conv.weight']
            new_checkpoint[
                f'decoder.up_blocks.{i}.upsamplers.0.conv.bias'] = \
                vae_state_dict[f'decoder.up.{block_id}.upsample.conv.bias']

        paths = renew_vae_resnet_paths(resnets)
        meta_path = {
            'old': f'up.{block_id}.block',
            'new': f'up_blocks.{i}.resnets'
        }
        assign_to_checkpoint(
            paths,
            new_checkpoint,
            vae_state_dict,
            additional_replacements=[meta_path],
            config=config)

    mid_resnets = [key for key in vae_state_dict if 'decoder.mid.block' in key]
    num_mid_res_blocks = 2
    for i in range(1, num_mid_res_blocks + 1):
        resnets = [
            key for key in mid_resnets if f'decoder.mid.block_{i}' in key
        ]

        paths = renew_vae_resnet_paths(resnets)
        meta_path = {
            'old': f'mid.block_{i}',
            'new': f'mid_block.resnets.{i - 1}'
        }
        assign_to_checkpoint(
            paths,
            new_checkpoint,
            vae_state_dict,
            additional_replacements=[meta_path],
            config=config)

    mid_attentions = [
        key for key in vae_state_dict if 'decoder.mid.attn' in key
    ]
    paths = renew_vae_attention_paths(mid_attentions)
    meta_path = {'old': 'mid.attn_1', 'new': 'mid_block.attentions.0'}
    assign_to_checkpoint(
        paths,
        new_checkpoint,
        vae_state_dict,
        additional_replacements=[meta_path],
        config=config)
    conv_attn_to_linear(new_checkpoint)

    new_checkpoint = {
        'model.' + key: value
        for key, value in new_checkpoint.items()
    }
    new_checkpoint = {
        key.replace('query', 'to_q'): value
        for key, value in new_checkpoint.items()
    }
    new_checkpoint = {
        key.replace('key', 'to_k'): value
        for key, value in new_checkpoint.items()
    }
    new_checkpoint = {
        key.replace('value', 'to_v'): value
        for key, value in new_checkpoint.items()
    }
    new_checkpoint = {
        key.replace('proj_attn', 'to_out.0'): value
        for key, value in new_checkpoint.items()
    }
    return new_checkpoint


def convert_ldm_clip_checkpoint(checkpoint):
    text_model = CLIPTextModel.from_pretrained('openai/clip-vit-large-patch14')
    keys = list(checkpoint.keys())

    text_model_dict = {}

    for key in keys:
        if key.startswith('cond_stage_model.transformer'):
            text_model_dict[
                key[len('cond_stage_model.transformer.'):]] = checkpoint[key]
    # Certain text transformers no longer
    # expect position_ids after transformers==4.31
    position_id_key = 'text_model.embeddings.position_ids'
    if position_id_key in text_model_dict and \
            position_id_key not in text_model.state_dict():
        del text_model_dict[position_id_key]

    text_model.load_state_dict(text_model_dict)

    return text_model


textenc_conversion_lst = [
    ('cond_stage_model.model.positional_embedding',
     'text_model.embeddings.position_embedding.weight'),
    ('cond_stage_model.model.token_embedding.weight',
     'text_model.embeddings.token_embedding.weight'),
    ('cond_stage_model.model.ln_final.weight',
     'text_model.final_layer_norm.weight'),
    ('cond_stage_model.model.ln_final.bias',
     'text_model.final_layer_norm.bias'),
]
textenc_conversion_map = {x[0]: x[1] for x in textenc_conversion_lst}

textenc_transformer_conversion_lst = [
    # (stable-diffusion, HF Diffusers)
    ('resblocks.', 'text_model.encoder.layers.'),
    ('ln_1', 'layer_norm1'),
    ('ln_2', 'layer_norm2'),
    ('.c_fc.', '.fc1.'),
    ('.c_proj.', '.fc2.'),
    ('.attn', '.self_attn'),
    ('ln_final.', 'transformer.text_model.final_layer_norm.'),
    ('token_embedding.weight',
     'transformer.text_model.embeddings.token_embedding.weight'),
    ('positional_embedding',
     'transformer.text_model.embeddings.position_embedding.weight'),
]
protected = {re.escape(x[0]): x[1] for x in textenc_transformer_conversion_lst}
textenc_pattern = re.compile('|'.join(protected.keys()))


def convert_paint_by_example_checkpoint(checkpoint):
    config = CLIPVisionConfig.from_pretrained('openai/clip-vit-large-patch14')
    model = PaintByExampleImageEncoder(config)

    keys = list(checkpoint.keys())

    text_model_dict = {}

    for key in keys:
        if key.startswith('cond_stage_model.transformer'):
            text_model_dict[
                key[len('cond_stage_model.transformer.'):]] = checkpoint[key]

    # load clip vision
    model.model.load_state_dict(text_model_dict)

    # load mapper
    keys_mapper = {
        k[len('cond_stage_model.mapper.res'):]: v
        for k, v in checkpoint.items()
        if k.startswith('cond_stage_model.mapper')
    }

    MAPPING = {
        'attn.c_qkv': ['attn1.to_q', 'attn1.to_k', 'attn1.to_v'],
        'attn.c_proj': ['attn1.to_out.0'],
        'ln_1': ['norm1'],
        'ln_2': ['norm3'],
        'mlp.c_fc': ['ff.net.0.proj'],
        'mlp.c_proj': ['ff.net.2'],
    }

    mapped_weights = {}
    for key, value in keys_mapper.items():
        prefix = key[:len('blocks.i')]
        suffix = key.split(prefix)[-1].split('.')[-1]
        name = key.split(prefix)[-1].split(suffix)[0][1:-1]
        mapped_names = MAPPING[name]

        num_splits = len(mapped_names)
        for i, mapped_name in enumerate(mapped_names):
            new_name = '.'.join([prefix, mapped_name, suffix])
            shape = value.shape[0] // num_splits
            mapped_weights[new_name] = value[i * shape:(i + 1) * shape]

    model.mapper.load_state_dict(mapped_weights)

    # load final layer norm
    model.final_layer_norm.load_state_dict({
        'bias':
        checkpoint['cond_stage_model.final_ln.bias'],
        'weight':
        checkpoint['cond_stage_model.final_ln.weight'],
    })

    # load final proj
    model.proj_out.load_state_dict({
        'bias': checkpoint['proj_out.bias'],
        'weight': checkpoint['proj_out.weight'],
    })

    # load uncond vector
    model.uncond_vector.data = torch.nn.Parameter(
        checkpoint['learnable_vector'])
    return model


def convert_open_clip_checkpoint(checkpoint):
    text_model = CLIPTextModel.from_pretrained(
        'stabilityai/stable-diffusion-2', subfolder='text_encoder')

    keys = list(checkpoint.keys())

    text_model_dict = {}

    if 'cond_stage_model.model.text_projection' in checkpoint:
        d_model = int(
            checkpoint['cond_stage_model.model.text_projection'].shape[0])
    else:
        d_model = 1024

    text_model_dict[
        'text_model.embeddings.position_ids'] = \
        text_model.text_model.embeddings.get_buffer('position_ids')

    for key in keys:
        # Diffusers drops the final layer and
        # only uses the penultimate layer
        if 'resblocks.23' in key:
            continue
        if key in textenc_conversion_map:
            text_model_dict[textenc_conversion_map[key]] = checkpoint[key]
        if key.startswith('cond_stage_model.model.transformer.'):
            new_key = key[len('cond_stage_model.model.transformer.'):]
            if new_key.endswith('.in_proj_weight'):
                new_key = new_key[:-len('.in_proj_weight')]
                new_key = textenc_pattern.sub(
                    lambda m: protected[re.escape(m.group(0))], new_key)
                text_model_dict[
                    new_key + '.q_proj.weight'] = checkpoint[key][:d_model, :]
                text_model_dict[new_key + '.k_proj.weight'] = checkpoint[key][
                    d_model:d_model * 2, :]
                text_model_dict[new_key +
                                '.v_proj.weight'] = checkpoint[key][d_model *
                                                                    2:, :]
            elif new_key.endswith('.in_proj_bias'):
                new_key = new_key[:-len('.in_proj_bias')]
                new_key = textenc_pattern.sub(
                    lambda m: protected[re.escape(m.group(0))], new_key)
                text_model_dict[new_key +
                                '.q_proj.bias'] = checkpoint[key][:d_model]
                text_model_dict[
                    new_key +
                    '.k_proj.bias'] = checkpoint[key][d_model:d_model * 2]
                text_model_dict[new_key +
                                '.v_proj.bias'] = checkpoint[key][d_model * 2:]
            else:
                new_key = textenc_pattern.sub(
                    lambda m: protected[re.escape(m.group(0))], new_key)

                text_model_dict[new_key] = checkpoint[key]

    text_model.load_state_dict(text_model_dict)

    return text_model


def stable_unclip_image_encoder(original_config):
    """Returns the image processor and clip image encoder for the img2img
    unclip pipeline.

    We currently know of two types of stable unclip models which separately use
    the clip and the openclip image encoders.
    """

    image_embedder_config = original_config.model.params.embedder_config

    sd_clip_image_embedder_class = image_embedder_config.target
    sd_clip_image_embedder_class = sd_clip_image_embedder_class.split('.')[-1]

    if sd_clip_image_embedder_class == 'ClipImageEmbedder':
        clip_model_name = image_embedder_config.params.model

        if clip_model_name == 'ViT-L/14':
            feature_extractor = CLIPImageProcessor()
            image_encoder = CLIPVisionModelWithProjection.from_pretrained(
                'openai/clip-vit-large-patch14')
        else:
            raise NotImplementedError(
                f'Unknown CLIP checkpoint name in stable '
                f'diffusion checkpoint {clip_model_name}')

    elif sd_clip_image_embedder_class == 'FrozenOpenCLIPImageEmbedder':
        feature_extractor = CLIPImageProcessor()
        image_encoder = CLIPVisionModelWithProjection.from_pretrained(
            'laion/CLIP-ViT-H-14-laion2B-s32B-b79K')
    else:
        raise NotImplementedError(
            f'Unknown CLIP image embedder class in '
            f'stable diffusion checkpoint {sd_clip_image_embedder_class}')

    return feature_extractor, image_encoder


def stable_unclip_image_noising_components(
        original_config,
        clip_stats_path: Optional[str] = None,
        device: Optional[str] = None):
    """Returns the noising components for the img2img and txt2img unclip
    pipelines.

    Converts the stability noise augmentor into
    1. a `StableUnCLIPImageNormalizer` for holding the CLIP stats
    2. a `DDPMScheduler` for holding the noise schedule

    If the noise augmentor config specifies a clip stats path,
    the `clip_stats_path` must be provided.
    """
    noise_aug_config = original_config.model.params.noise_aug_config
    noise_aug_class = noise_aug_config.target
    noise_aug_class = noise_aug_class.split('.')[-1]

    if noise_aug_class == 'CLIPEmbeddingNoiseAugmentation':
        noise_aug_config = noise_aug_config.params
        embedding_dim = noise_aug_config.timestep_dim
        max_noise_level = noise_aug_config.noise_schedule_config.timesteps
        beta_schedule = noise_aug_config.noise_schedule_config.beta_schedule

        image_normalizer = StableUnCLIPImageNormalizer(
            embedding_dim=embedding_dim)
        image_noising_scheduler = DDPMScheduler(
            num_train_timesteps=max_noise_level, beta_schedule=beta_schedule)

        if 'clip_stats_path' in noise_aug_config:
            if clip_stats_path is None:
                raise ValueError(
                    'This stable unclip config requires a `clip_stats_path`')

            clip_mean, clip_std = torch.load(
                clip_stats_path, map_location=device)
            clip_mean = clip_mean[None, :]
            clip_std = clip_std[None, :]

            clip_stats_state_dict = {
                'mean': clip_mean,
                'std': clip_std,
            }

            image_normalizer.load_state_dict(clip_stats_state_dict)
    else:
        raise NotImplementedError(
            f'Unknown noise augmentor class: {noise_aug_class}')

    return image_normalizer, image_noising_scheduler


def save_videos_grid(videos: torch.Tensor,
                     path: str,
                     rescale=False,
                     n_rows=6,
                     fps=8):
    videos = rearrange(videos, 'b c t h w -> t b c h w')
    outputs = []
    for x in videos:
        x = torchvision.utils.make_grid(x, nrow=n_rows)
        x = x.transpose(0, 1).transpose(1, 2).squeeze(-1)
        if rescale:
            x = (x + 1.0) / 2.0  # -1,1 -> 0,1
        x = (x * 255).numpy().astype(np.uint8)
        outputs.append(x)

    os.makedirs(os.path.dirname(path), exist_ok=True)
    # imageio v3 doesn't support fps
    if imageio.__version__ < '2.28.0':
        imageio.mimsave(path, outputs, fps=fps)
    else:
        imageio.mimsave(path, outputs, duration=1000 * 1 / fps, loop=10)



# Copyright (c) OpenMMLab. All rights reserved.
from .stable_diffusion import StableDiffusion
from .stable_diffusion_inpaint import StableDiffusionInpaint
from .vae import AutoencoderKL

__all__ = ['StableDiffusion', 'AutoencoderKL', 'StableDiffusionInpaint']



# Copyright (c) OpenMMLab. All rights reserved.
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
from mmengine.model import BaseModule

from mmagic.registry import MODELS
from ..pggan import (EqualizedLRConvDownModule, EqualizedLRConvModule,
                     MiniBatchStddevLayer)
from .stylegan1_modules import Blur, EqualLinearActModule


@MODELS.register_module('StyleGANv1Discriminator')
@MODELS.register_module()
class StyleGAN1Discriminator(BaseModule):
    """StyleGAN1 Discriminator.

    The architecture of this discriminator is proposed in StyleGAN1. More
    details can be found in: A Style-Based Generator Architecture for
    Generative Adversarial Networks CVPR2019.

    Args:
        in_size (int): The input size of images.
        blur_kernel (list, optional): The blurry kernel. Defaults
            to [1, 2, 1].
        mbstd_cfg (dict, optional): Configs for minibatch-stddev layer.
            Defaults to dict(group_size=4).
    """

    def __init__(self,
                 in_size,
                 blur_kernel=[1, 2, 1],
                 mbstd_cfg=dict(group_size=4)):
        super().__init__()

        self.with_mbstd = mbstd_cfg is not None
        channels = {
            4: 512,
            8: 512,
            16: 512,
            32: 512,
            64: 256,
            128: 128,
            256: 64,
            512: 32,
            1024: 16,
        }

        log_size = int(np.log2(in_size))
        self.log_size = log_size
        in_channels = channels[in_size]

        self.convs = nn.ModuleList()
        self.from_rgb = nn.ModuleList()

        for i in range(log_size, 2, -1):
            out_channel = channels[2**(i - 1)]
            self.from_rgb.append(
                EqualizedLRConvModule(
                    3,
                    in_channels,
                    kernel_size=3,
                    padding=1,
                    act_cfg=dict(type='LeakyReLU', negative_slope=0.2)))
            self.convs.append(
                nn.Sequential(
                    EqualizedLRConvModule(
                        in_channels,
                        out_channel,
                        kernel_size=3,
                        padding=1,
                        bias=True,
                        norm_cfg=None,
                        act_cfg=dict(type='LeakyReLU', negative_slope=0.2)),
                    Blur(blur_kernel, pad=(1, 1)),
                    EqualizedLRConvDownModule(
                        out_channel,
                        out_channel,
                        kernel_size=3,
                        stride=2,
                        padding=1,
                        act_cfg=None),
                    nn.LeakyReLU(negative_slope=0.2, inplace=True)))

            in_channels = out_channel

        self.from_rgb.append(
            EqualizedLRConvModule(
                3,
                in_channels,
                kernel_size=3,
                padding=0,
                act_cfg=dict(type='LeakyReLU', negative_slope=0.2)))
        self.convs.append(
            nn.Sequential(
                EqualizedLRConvModule(
                    in_channels + 1,
                    512,
                    kernel_size=3,
                    padding=1,
                    bias=True,
                    norm_cfg=None,
                    act_cfg=dict(type='LeakyReLU', negative_slope=0.2)),
                EqualizedLRConvModule(
                    512,
                    512,
                    kernel_size=4,
                    padding=0,
                    bias=True,
                    norm_cfg=None,
                    act_cfg=None),
            ))

        if self.with_mbstd:
            self.mbstd_layer = MiniBatchStddevLayer(**mbstd_cfg)

        self.final_linear = nn.Sequential(EqualLinearActModule(channels[4], 1))

        self.n_layer = len(self.convs)

    def forward(self, input, transition_weight=1., curr_scale=-1):
        """Forward function.

        Args:
            input (torch.Tensor): Input image tensor.
            transition_weight (float, optional): The weight used in resolution
                transition. Defaults to 1..
            curr_scale (int, optional): The resolution scale of image tensor.
                -1 means the max resolution scale of the StyleGAN1.
                Defaults to -1.

        Returns:
            torch.Tensor: Predict score for the input image.
        """
        curr_log_size = self.log_size if curr_scale < 0 else int(
            np.log2(curr_scale))
        step = curr_log_size - 2
        for i in range(step, -1, -1):
            index = self.n_layer - i - 1

            if i == step:
                out = self.from_rgb[index](input)

            # minibatch standard deviation
            if i == 0:
                out = self.mbstd_layer(out)

            out = self.convs[index](out)

            if i > 0:
                if i == step and 0 <= transition_weight < 1:
                    skip_rgb = F.avg_pool2d(input, 2)
                    skip_rgb = self.from_rgb[index + 1](skip_rgb)

                    out = (1 - transition_weight
                           ) * skip_rgb + transition_weight * out

        out = out.view(out.shape[0], -1)
        out = self.final_linear(out)
        return out



# Copyright (c) OpenMMLab. All rights reserved.
import random
from collections import deque

import torch
import torch.nn as nn

from mmagic.registry import MODELS
from .gan_loss import GANLoss, gradient_penalty_loss


class ImagePool:
    """Defined a image pool for RelativisticDiscLoss."""

    def __init__(self, pool_size):
        self.pool_size = pool_size
        self.sample_size = pool_size
        if self.pool_size > 0:
            self.num_imgs = 0
            self.images = deque()

    def add(self, images):
        if self.pool_size == 0:
            return images
        for image in images.data:
            image = torch.unsqueeze(image, 0)
            if self.num_imgs < self.pool_size:
                self.num_imgs = self.num_imgs + 1
                self.images.append(image)
            else:
                self.images.popleft()
                self.images.append(image)

    def query(self):
        if len(self.images) > self.sample_size:
            return_images = list(random.sample(self.images, self.sample_size))
        else:
            return_images = list(self.images)
        return torch.cat(return_images, 0)


class DiscLoss(nn.Module):
    """Defined a criterion to calculator loss."""

    def name(self):
        """return name of criterion."""
        return 'DiscLoss'

    def __init__(self, gan_type='vanilla'):
        super(DiscLoss, self).__init__()

        self.criterionGAN = GANLoss(gan_type=gan_type)

    def forward(self, net, fakeB, realB, model='discriminator'):
        """Get discriminator or generator loss."""
        if model == 'discriminator':
            self.pred_fake = net.forward(fakeB.detach())
            self.loss_D_fake = self.criterionGAN(self.pred_fake, 0)

            # Real
            self.pred_real = net.forward(realB)
            self.loss_D_real = self.criterionGAN(self.pred_real, 1)

            # Combined loss
            self.loss_D = (self.loss_D_fake + self.loss_D_real) * 0.5
            return self.loss_D
        else:
            pred_fake = net.forward(fakeB)
            return self.criterionGAN(pred_fake, 1)


class RelativisticDiscLoss(nn.Module):
    """Defined a criterion to calculator loss."""

    def name(self):
        """return name of criterion."""
        return 'RelativisticDiscLoss'

    def __init__(self):
        super(RelativisticDiscLoss, self).__init__()

        self.criterionGAN = GANLoss(gan_type='vanilla')
        self.fake_pool = ImagePool(50)
        self.real_pool = ImagePool(50)

    def forward(self, net, fakeB, realB, model='discriminator'):
        """Get discriminator or generator loss."""
        if model == 'discriminator':
            self.fake_B = fakeB.detach()
            self.real_B = realB
            self.pred_fake = net.forward(self.fake_B)
            self.fake_pool.add(self.pred_fake)

            # Real
            self.pred_real = net.forward(self.real_B)
            self.real_pool.add(self.pred_real)

            # Combined loss
            self.loss_D = (self.criterionGAN(
                self.pred_real - torch.mean(self.fake_pool.query()), 1) +
                           self.criterionGAN(
                               self.pred_fake -
                               torch.mean(self.real_pool.query()), 0)) / 2
            return self.loss_D
        else:
            self.pred_fake = net.forward(fakeB)

            # Real
            self.pred_real = net.forward(realB)
            errG = (self.criterionGAN(
                self.pred_real - torch.mean(self.fake_pool.query()), 0) +
                    self.criterionGAN(
                        self.pred_fake - torch.mean(self.real_pool.query()),
                        1)) / 2
            return errG


class RelativisticDiscLossLS(nn.Module):
    """Defined a criterion to calculator loss."""

    def name(self):
        """return name of criterion."""
        return 'RelativisticDiscLossLS'

    def __init__(self):
        super(RelativisticDiscLossLS, self).__init__()

        self.criterionGAN = GANLoss(gan_type='l1')
        self.fake_pool = ImagePool(50)
        self.real_pool = ImagePool(50)

    def forward(self, net, fakeB, realB, model='discriminator'):
        """Get discriminator or generator loss."""
        if model == 'discriminator':
            self.fake_B = fakeB.detach()
            self.real_B = realB
            self.pred_fake = net.forward(fakeB.detach())
            self.fake_pool.add(self.pred_fake)

            # Real
            self.pred_real = net.forward(realB)
            self.real_pool.add(self.pred_real)

            # Combined loss
            ex_pdata = torch.mean(
                (self.pred_real - torch.mean(self.fake_pool.query()) - 1)**2)
            ex_pz = torch.mean(
                (self.pred_fake - torch.mean(self.real_pool.query()) + 1)**2)
            self.loss_D = (ex_pdata + ex_pz) / 2
            return self.loss_D
        else:
            self.pred_fake = net.forward(fakeB)

            # Real
            self.pred_real = net.forward(realB)
            ex_pdata = torch.mean(
                (self.pred_real - torch.mean(self.fake_pool.query()) + 1)**2)
            ez_pz = torch.mean(
                (self.pred_fake - torch.mean(self.real_pool.query()) - 1)**2)
            errG = (ex_pdata + ez_pz) / 2
            return errG


class DiscLossWGANGP(DiscLoss):
    """Defined a criterion to calculator loss."""

    def name(self):
        """return name of criterion."""
        return 'DiscLossWGAN-GP'

    def __init__(self):
        super(DiscLossWGANGP, self).__init__()
        self.LAMBDA = 10

    def forward(self, net, fakeB, realB, model='discriminator'):
        """Get discriminator or generator loss."""
        if model == 'discriminator':
            self.D_fake = net.forward(fakeB.detach())
            self.D_fake = self.D_fake.mean()

            # Real
            self.D_real = net.forward(realB)
            self.D_real = self.D_real.mean()
            # Combined loss
            self.loss_D = self.D_fake - self.D_real
            gradient_penalty = gradient_penalty_loss(net, realB.data,
                                                     fakeB.data) * self.LAMBDA
            return self.loss_D + gradient_penalty
        else:
            self.D_fake = net.forward(fakeB)
            return -self.D_fake.mean()


@MODELS.register_module()
class AdvLoss:
    """Returns the loss of discriminator with the specified type for
    DeblurGanv2.

    Args:
        loss_type (Str): One of value in [wgan-gp,lsgan,gan,ragan,ragan-ls].
    """

    def __new__(cls, loss_type: str):
        if loss_type == 'wgan-gp':
            disc_loss = DiscLossWGANGP()
        elif loss_type == 'lsgan':
            disc_loss = DiscLoss(gan_type='l1')
        elif loss_type == 'gan':
            disc_loss = DiscLoss()
        elif loss_type == 'ragan':
            disc_loss = RelativisticDiscLoss()
        elif loss_type == 'ragan-ls':
            disc_loss = RelativisticDiscLossLS()
        else:
            raise ValueError('GAN Loss [%s] not recognized.' % loss_type)
        return disc_loss



# Copyright (c) OpenMMLab. All rights reserved.
from collections import OrderedDict
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import torch
from PIL import Image
from torch import nn
from torch.hub import load_state_dict_from_url
from tqdm import tqdm
from transformers import CLIPTokenizer

from mmagic.registry import MODELS
from ..stable_diffusion import StableDiffusion
from .fastcomposer_util import FastComposerModel, get_object_transforms

ModelType = Union[Dict, nn.Module]


@MODELS.register_module()
class FastComposer(StableDiffusion):

    def __init__(self,
                 pretrained_cfg: dict,
                 vae: ModelType,
                 text_encoder: ModelType,
                 tokenizer: str,
                 unet: ModelType,
                 scheduler: ModelType,
                 test_scheduler: Optional[ModelType] = None,
                 dtype: str = 'fp32',
                 enable_xformers: bool = True,
                 noise_offset_weight: float = 0,
                 tomesd_cfg: Optional[dict] = None,
                 data_preprocessor=dict(type='DataPreprocessor'),
                 init_cfg: Optional[dict] = None):
        super().__init__(vae, text_encoder, tokenizer, unet, scheduler,
                         test_scheduler, dtype, enable_xformers,
                         noise_offset_weight, tomesd_cfg, data_preprocessor,
                         init_cfg)
        self.vae = self.vae.model
        self.unet = self.unet.model
        model = FastComposerModel.from_pretrained(pretrained_cfg, self.vae,
                                                  self.unet)
        if pretrained_cfg['finetuned_model_path'][:7] == 'http://' or \
                pretrained_cfg['finetuned_model_path'][:8] == 'https://':
            model.load_state_dict(
                load_state_dict_from_url(
                    pretrained_cfg['finetuned_model_path'],
                    map_location='cpu'),
                strict=False)
        elif pretrained_cfg['finetuned_model_path']:
            model.load_state_dict(
                torch.load(
                    pretrained_cfg['finetuned_model_path'],
                    map_location='cpu'),
                strict=False)
        weight_dtype = torch.float32
        if dtype == 'fp16':
            weight_dtype = torch.float16
        elif dtype == 'bf16':
            weight_dtype = torch.bfloat16
        model = model.to(weight_dtype)

        self.unet = model.unet

        if pretrained_cfg['enable_xformers_memory_efficient_attention']:
            self.unet.enable_xformers_memory_efficient_attention()

        self.text_encoder = model.text_encoder
        self.image_encoder = model.image_encoder

        self.postfuse_module = model.postfuse_module

        del model

        self.special_tokenizer = CLIPTokenizer.from_pretrained(
            pretrained_cfg['pretrained_model_name_or_path'],
            subfolder='tokenizer',
            revision=pretrained_cfg['revision'],
        )
        self.special_tokenizer.add_tokens(['img'], special_tokens=True)
        self.image_token_id = self.special_tokenizer.convert_tokens_to_ids(
            'img')

        self.object_transforms = get_object_transforms(pretrained_cfg)

    @torch.no_grad()
    def _tokenize_and_mask_noun_phrases_ends(self, caption):
        """Augment the text embedding."""

        input_ids = self.special_tokenizer.encode(caption)
        noun_phrase_end_mask = [False for _ in input_ids]
        clean_input_ids = []
        clean_index = 0

        for i, id in enumerate(input_ids):
            if id == self.image_token_id:
                noun_phrase_end_mask[clean_index - 1] = True
            else:
                clean_input_ids.append(id)
                clean_index += 1

        max_len = self.special_tokenizer.model_max_length

        if len(clean_input_ids) > max_len:
            clean_input_ids = clean_input_ids[:max_len]
        else:
            clean_input_ids = clean_input_ids + [
                self.tokenizer.pad_token_id
            ] * (
                max_len - len(clean_input_ids))

        if len(noun_phrase_end_mask) > max_len:
            noun_phrase_end_mask = noun_phrase_end_mask[:max_len]
        else:
            noun_phrase_end_mask = noun_phrase_end_mask + [False] * (
                max_len - len(noun_phrase_end_mask))

        clean_input_ids = torch.tensor(clean_input_ids, dtype=torch.long)
        noun_phrase_end_mask = torch.tensor(
            noun_phrase_end_mask, dtype=torch.bool)
        return clean_input_ids.unsqueeze(0), noun_phrase_end_mask.unsqueeze(0)

    @torch.no_grad()
    def _encode_augmented_prompt(self, prompt: str,
                                 reference_images: List[Image.Image],
                                 device: torch.device,
                                 weight_dtype: torch.dtype):
        """Encode reference images.

        Args:
            prompt (str or list(int)): prompt to be encoded.
            reference_images: (List[Image.Image]): List of reference images.
            device (torch.device):torch device.
            weight_dtype (torch.dtype): torch.dtype.

        Returns:
            text_embeddings (torch.Tensor): text embeddings generated by
                clip text encoder.
        """
        # TODO: check this
        object_pixel_values = []
        for image in reference_images:
            image_tensor = torch.from_numpy(np.array(
                image.convert('RGB'))).permute(2, 0, 1)
            image = self.object_transforms(image_tensor)
            object_pixel_values.append(image)

        object_pixel_values = torch.stack(
            object_pixel_values,
            dim=0).to(memory_format=torch.contiguous_format).float()
        object_pixel_values = object_pixel_values.unsqueeze(0).to(
            dtype=weight_dtype, device=device)
        object_embeds = self.image_encoder(object_pixel_values)

        id_and_mask = self._tokenize_and_mask_noun_phrases_ends(prompt)
        input_ids, image_token_mask = id_and_mask
        input_ids, image_token_mask = input_ids.to(
            device), image_token_mask.to(device)

        num_objects = image_token_mask.sum(dim=1)

        augmented_prompt_embeds = self.postfuse_module(
            self.text_encoder(input_ids)[0], object_embeds, image_token_mask,
            num_objects)
        return augmented_prompt_embeds

    @torch.no_grad()
    def infer(self,
              prompt: Union[str, List[str]] = None,
              height: Optional[int] = None,
              width: Optional[int] = None,
              num_inference_steps: int = 50,
              guidance_scale: float = 7.5,
              negative_prompt: Optional[Union[str, List[str]]] = None,
              num_images_per_prompt: Optional[int] = 1,
              eta: float = 0.0,
              generator: Optional[Union[torch.Generator,
                                        List[torch.Generator]]] = None,
              latents: Optional[torch.FloatTensor] = None,
              prompt_embeds: Optional[torch.FloatTensor] = None,
              negative_prompt_embeds: Optional[torch.FloatTensor] = None,
              output_type: Optional[str] = 'pil',
              return_dict: bool = True,
              callback: Optional[Callable[[int, int, torch.FloatTensor],
                                          None]] = None,
              callback_steps: int = 1,
              cross_attention_kwargs: Optional[Dict[str, Any]] = None,
              alpha_: float = 0.7,
              reference_subject_images: List[Image.Image] = None,
              augmented_prompt_embeds: Optional[torch.FloatTensor] = None,
              show_progress: bool = True):
        r"""
        Function invoked when calling the pipeline for generation.

        Args:
            prompt (`str` or `List[str]`, *optional*):
                The prompt or prompts to guide the image generation.
                If not defined, one has to pass `prompt_embeds`.
                instead.
            height (`int`, *optional*):
                defaults to
                self.unet.config.sample_size * self.vae_scale_factor
                The height in pixels of the generated image.
            width (`int`, *optional*):
                defaults to
                self.unet.config.sample_size * self.vae_scale_factor
                The width in pixels of the generated image.
            num_inference_steps (`int`, *optional*, defaults to 50):
                The number of denoising steps.
                More denoising steps usually lead to a higher quality image
                at the expense of slower inference.
            guidance_scale (`float`, *optional*, defaults to 7.5):
                Guidance scale as defined in
                [Classifier-Free Diffusion Guidance]
                (https://arxiv.org/abs/2207.12598).
                `guidance_scale` is defined as `w` of equation
                2. of [Imagen Paper](https://arxiv.org/pdf/2205.11487.pdf).
                Guidance scale is enabled by setting `guidance_scale >
                1`. Higher guidance scale encourages to generate images
                that are closely linked to the text `prompt`,
                usually at the expense of lower image quality.
            negative_prompt (`str` or `List[str]`, *optional*):
                The prompt or prompts not to guide the image generation.
                If not defined, one has to pass
                `negative_prompt_embeds` instead.
                Ignored when not using guidance
                (i.e., ignored if `guidance_scale` is less than `1`).
            num_images_per_prompt (`int`, *optional*, defaults to 1):
                The number of images to generate per prompt.
            eta (`float`, *optional*, defaults to 0.0):
                Corresponds to parameter eta (η) in the DDIM paper:
                https://arxiv.org/abs/2010.02502. Only applies to
                [`schedulers.DDIMScheduler`], will be ignored for others.
            generator (`torch.Generator` or `List[torch.Generator]`,
             *optional*):
                One or a list of [torch generator(s)](https://pytorch.org/
                docs/stable/generated/torch.Generator.html)
                to make generation deterministic.
            latents (`torch.FloatTensor`, *optional*):
                Pre-generated noisy latents,
                sampled from a Gaussian distribution,
                to be used as inputs for image generation.
                Can be used to tweak the same generation with
                different prompts. If not provided, a latents tensor
                will ge generated by sampling using
                the supplied random `generator`.
            prompt_embeds (`torch.FloatTensor`, *optional*):
                Pre-generated text embeddings.
                Can be used to easily tweak text inputs,
                *e.g.* prompt weighting. If not provided,
                text embeddings will be generated from `prompt`
                input argument.
            negative_prompt_embeds (`torch.FloatTensor`, *optional*):
                Pre-generated negative text embeddings.
                Can be used to easily tweak text inputs, *e.g.* prompt
                weighting. If not provided, negative_prompt_embeds will be
                generated from `negative_prompt` input argument.
            output_type (`str`, *optional*, defaults to `"pil"`):
                The output format of the generate image. Choose between
                [PIL](https://pillow.readthedocs.io/en/stable/):
                `PIL.Image.Image` or `np.array`.
            return_dict (`bool`, *optional*, defaults to `True`):
                Whether or not to return a
                [`~pipelines.stable_diffusion.StableDiffusionPipelineOutput`]
                instead of a plain tuple.
            callback (`Callable`, *optional*):
                A function that will be called every `callback_steps`
                steps during inference. The function will be
                called with the following arguments: `callback(step: int,
                timestep: int, latents: torch.FloatTensor)`.
            callback_steps (`int`, *optional*, defaults to 1):
                The frequency at which the `callback` function will be called.
                If not specified, the callback will be called at every step.
            cross_attention_kwargs (`dict`, *optional*):
                A kwargs dictionary that if specified is passed along to the
                `AttentionProcessor` as defined under `self.processor` in
                [diffusers.cross_attention](https://github.com/huggingface/
                diffusers/blob/main/src/diffusers/models/cross_attention.py).
            alpha_ (`float`, defaults to 0.7):
                The ratio of subject conditioning. If `alpha_` is 0.7,
                the beginning 30% of denoising steps use text prompts,
                while the last 70% utilize image-augmented prompts.
                Increase alpha for identity preservation,
                decrease it for prompt consistency.
            reference_subject_images (`List[PIL.Image.Image]`):
                a list of PIL images that are used as reference subjects.
                The number of images should be equal to the
                number of augmented tokens in the prompts.
            augmented_prompt_embeds: (`torch.FloatTensor`, *optional*):
                Pre-generated image augmented text embeddings.
                If not provided, embeddings will be generated from `prompt`
                and `reference_subject_images`.
            show_progress: ('bool'):
                show progress or not.
        Examples:

        Returns:
            `OrderedDict` or `tuple`:
            `OrderedDict` if `return_dict` is True, otherwise a `tuple.
            When returning a tuple, the first element is a list with
            the generated images, and the second element is a
            list of `bool`s denoting whether the corresponding
            generated image likely represents "not-safe-for-work"
            (nsfw) content, according to the `safety_checker`.
        """
        # 0. Default height and width to unet
        height = height or self.unet.config.sample_size * self.vae_scale_factor
        width = width or self.unet.config.sample_size * self.vae_scale_factor

        # 1. Check inputs. Raise error if not correct
        self.check_inputs(
            prompt,
            height,
            width,
        )

        assert_text = 'Prompt and reference subject images or prompt_embeds ' \
                      'and augmented_prompt_embeds must be provided.'

        assert (prompt and reference_subject_images) or \
               (prompt_embeds and augmented_prompt_embeds), assert_text

        # 2. Define call parameters
        if prompt is not None and isinstance(prompt, str):
            batch_size = 1
        elif prompt is not None and isinstance(prompt, list):
            batch_size = len(prompt)
        else:
            batch_size = prompt_embeds.shape[0]

        device = self.device
        # here `guidance_scale` is defined analog to
        # the guidance weight `w` of equation (2)
        # of the Imagen paper: https://arxiv.org/pdf/2205.11487.pdf .
        # `guidance_scale = 1`
        # corresponds to doing no classifier free guidance.
        do_classifier_free_guidance = guidance_scale > 1.0

        assert do_classifier_free_guidance

        # 3. Encode input prompt
        prompt_text_only = prompt.replace('img', '')

        prompt_embeds = self._encode_prompt(prompt_text_only, device,
                                            num_images_per_prompt,
                                            do_classifier_free_guidance,
                                            negative_prompt)

        if augmented_prompt_embeds is None:
            augmented_prompt_embeds = self._encode_augmented_prompt(
                prompt, reference_subject_images, device, prompt_embeds.dtype)
            augmented_prompt_embeds = augmented_prompt_embeds.repeat(
                num_images_per_prompt, 1, 1)

        prompt_embeds = torch.cat([prompt_embeds, augmented_prompt_embeds],
                                  dim=0)

        # 4. Prepare timesteps
        self.scheduler.set_timesteps(num_inference_steps, device=device)
        timesteps = self.scheduler.timesteps

        # 5. Prepare latent variables
        num_channels_latents = self.unet.in_channels
        latents = self.prepare_latents(
            batch_size * num_images_per_prompt,
            num_channels_latents,
            height,
            width,
            prompt_embeds.dtype,
            device,
            generator,
            latents,
        )

        start_subject_conditioning_step = (1 - alpha_) * num_inference_steps

        extra_step_kwargs = self.prepare_extra_step_kwargs(generator, eta)
        (null_prompt_embeds, text_prompt_embeds,
         augmented_prompt_embeds) = prompt_embeds.chunk(3)

        # 7. Denoising loop
        num_warmup_steps = len(
            timesteps) - num_inference_steps * self.scheduler.order
        if show_progress:
            timesteps = tqdm(timesteps)
        for i, t in enumerate(timesteps):
            latent_model_input = (
                torch.cat([latents] *
                          2) if do_classifier_free_guidance else latents)
            latent_model_input = self.scheduler.scale_model_input(
                latent_model_input, t)

            if i <= start_subject_conditioning_step:
                current_prompt_embeds = torch.cat(
                    [null_prompt_embeds, text_prompt_embeds], dim=0)
            else:
                current_prompt_embeds = torch.cat(
                    [null_prompt_embeds, augmented_prompt_embeds], dim=0)

            # predict the noise residual
            noise_pred = self.unet(
                latent_model_input,
                t,
                encoder_hidden_states=current_prompt_embeds,
                cross_attention_kwargs=cross_attention_kwargs,
            ).sample

            # perform guidance
            if do_classifier_free_guidance:
                noise_pred_uncond, noise_pred_text = noise_pred.chunk(2)
                noise_pred = noise_pred_uncond + guidance_scale * (
                    noise_pred_text - noise_pred_uncond)
            else:
                assert 0, 'Not Implemented'

            # compute the previous noisy sample x_t -> x_t-1
            latents = self.scheduler.step(noise_pred, t, latents,
                                          **extra_step_kwargs).prev_sample

            # call the callback, if provided
            if i == len(timesteps) - 1 or (i + 1 > num_warmup_steps and
                                           (i + 1) % self.scheduler.order
                                           == 0):
                if callback is not None and i % callback_steps == 0:
                    callback(i, t, latents)
        has_nsfw_concept = None
        if output_type == 'latent':
            image = latents
            has_nsfw_concept = None
        elif output_type == 'pil':
            # 8. Post-processing
            image = self.decode_latents(latents)

            # 10. Convert to PIL
            image = self.output_to_pil(image)
        else:
            # 8. Post-processing
            image = self.decode_latents(latents)

        # Offload last model to CPU
        if hasattr(
                self,
                'final_offload_hook') and self.final_offload_hook is not None:
            self.final_offload_hook.offload()

        if not return_dict:
            return (image, has_nsfw_concept)

        return OrderedDict(
            samples=image, nsfw_content_detected=has_nsfw_concept)



# Copyright (c) OpenMMLab. All rights reserved.
from .sagan import SAGAN
from .sagan_discriminator import ProjDiscriminator
from .sagan_generator import SNGANGenerator
from .sagan_modules import (SNGANDiscHeadResBlock, SNGANDiscResBlock,
                            SNGANGenResBlock)

__all__ = [
    'SAGAN', 'SNGANGenerator', 'ProjDiscriminator', 'SNGANDiscHeadResBlock',
    'SNGANDiscResBlock', 'SNGANGenResBlock'
]



# Copyright (c) OpenMMLab. All rights reserved.
from .cli import modify_args
from .img_utils import (all_to_tensor, can_convert_to_image, get_box_info,
                        reorder_image, tensor2img, to_numpy)
from .io_utils import MMAGIC_CACHE_DIR, download_from_url
# TODO replace with engine's API
from .logger import print_colored_log
from .sampler import get_sampler
from .setup_env import register_all_modules, try_import
from .trans_utils import (add_gaussian_noise, adjust_gamma, bbox2mask,
                          brush_stroke_mask, get_irregular_mask, make_coord,
                          random_bbox, random_choose_unknown)
from .typing import ConfigType, ForwardInputs, LabelVar, NoiseVar, SampleList

__all__ = [
    'modify_args', 'print_colored_log', 'register_all_modules', 'try_import',
    'ForwardInputs', 'SampleList', 'NoiseVar', 'ConfigType', 'LabelVar',
    'MMAGIC_CACHE_DIR', 'download_from_url', 'get_sampler', 'tensor2img',
    'random_choose_unknown', 'add_gaussian_noise', 'adjust_gamma',
    'make_coord', 'bbox2mask', 'brush_stroke_mask', 'get_irregular_mask',
    'random_bbox', 'reorder_image', 'to_numpy', 'get_box_info',
    'can_convert_to_image', 'all_to_tensor'
]



# Copyright (c) OpenMMLab. All rights reserved.
import torch
from mmengine.model import BaseModule
from torchvision import models

from mmagic.models.archs import ImgNormalize
from mmagic.registry import MODELS


@MODELS.register_module()
class LTE(BaseModule):
    """Learnable Texture Extractor.

    Based on pretrained VGG19. Generate features in 3 levels.

    Args:
        requires_grad (bool): Require grad or not. Default: True.
        pixel_range (float): Pixel range of feature. Default: 1.
        load_pretrained_vgg (bool): Load pretrained VGG from torchvision.
            Default: True.
            Train: must load pretrained VGG.
            Eval: needn't load pretrained VGG, because we will load pretrained
            LTE.
        init_cfg (dict, optional): Initialization config dict.
    """

    def __init__(self,
                 requires_grad=True,
                 pixel_range=1.,
                 load_pretrained_vgg=True,
                 init_cfg=None):
        super().__init__(init_cfg=init_cfg)

        vgg_mean = (0.485, 0.456, 0.406)
        vgg_std = (0.229 * pixel_range, 0.224 * pixel_range,
                   0.225 * pixel_range)
        self.img_normalize = ImgNormalize(
            pixel_range=pixel_range, img_mean=vgg_mean, img_std=vgg_std)

        # use vgg19 weights to initialize
        vgg_pretrained_features = models.vgg19(
            pretrained=load_pretrained_vgg).features

        self.slice1 = torch.nn.Sequential()
        self.slice2 = torch.nn.Sequential()
        self.slice3 = torch.nn.Sequential()

        for x in range(2):
            self.slice1.add_module(str(x), vgg_pretrained_features[x])
        for x in range(2, 7):
            self.slice2.add_module(str(x), vgg_pretrained_features[x])
        for x in range(7, 12):
            self.slice3.add_module(str(x), vgg_pretrained_features[x])

        if not requires_grad:
            for param in self.slice1.parameters():
                param.requires_grad = requires_grad
            for param in self.slice2.parameters():
                param.requires_grad = requires_grad
            for param in self.slice3.parameters():
                param.requires_grad = requires_grad

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (n, 3, h, w).

        Returns:
            Tuple[Tensor]: Forward results in 3 levels.
                x_level3: Forward results in level 3 (n, 256, h/4, w/4).
                x_level2: Forward results in level 2 (n, 128, h/2, w/2).
                x_level1: Forward results in level 1 (n, 64, h, w).
        """

        x = self.img_normalize(x)

        x_level1 = x = self.slice1(x)
        x_level2 = x = self.slice2(x)
        x_level3 = x = self.slice3(x)

        return [x_level3, x_level2, x_level1]



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os
import os.path as osp

import cv2
import numpy as np
import scipy.io as sio
import torch
from skimage import img_as_ubyte
from tqdm import tqdm


def export_images(args):
    """Export images from mat file."""

    noisy_dir = osp.join(args.out_dir, 'noisy')
    os.makedirs(noisy_dir, exist_ok=True)
    gt_dir = osp.join(args.out_dir, 'gt')
    os.makedirs(gt_dir, exist_ok=True)

    noisy_matfile = osp.join(args.data_root, 'ValidationNoisyBlocksSrgb.mat')
    gt_matfile = osp.join(args.data_root, 'ValidationGtBlocksSrgb.mat')
    noisy_images = sio.loadmat(noisy_matfile)
    gt_images = sio.loadmat(gt_matfile)

    noisy_images = np.float32(
        np.array(noisy_images['ValidationNoisyBlocksSrgb']))
    noisy_images /= 255.
    gt_images = np.float32(np.array(gt_images['ValidationGtBlocksSrgb']))
    gt_images /= 255.

    # processing noisy images
    print('Exporting', noisy_matfile, 'to', noisy_dir)
    for i in tqdm(range(40)):
        for k in range(32):
            noisy_patch = torch.from_numpy(
                noisy_images[i, k, :, :, :]).unsqueeze(0).permute(0, 3, 1, 2)
            noisy_patch = torch.clamp(noisy_patch, 0,
                                      1).detach().permute(0, 2, 3,
                                                          1).squeeze(0)
            save_path = osp.join(noisy_dir, f'val_{str(i*32+k)}_NOISY.png')
            cv2.imwrite(
                save_path,
                cv2.cvtColor(img_as_ubyte(noisy_patch), cv2.COLOR_RGB2BGR))

    # processing gt images
    print('Exporting', gt_matfile, 'to', gt_dir)
    for i in tqdm(range(40)):
        for k in range(32):
            gt_patch = torch.from_numpy(
                gt_images[i, k, :, :, :]).unsqueeze(0).permute(0, 3, 1, 2)
            gt_patch = torch.clamp(gt_patch, 0,
                                   1).detach().permute(0, 2, 3, 1).squeeze(0)
            save_path = osp.join(gt_dir, f'val_{str(i*32+k)}_GT.png')
            cv2.imwrite(
                save_path,
                cv2.cvtColor(img_as_ubyte(gt_patch), cv2.COLOR_RGB2BGR))

    print('\nFinish exporting images.')


def parse_args():
    parser = argparse.ArgumentParser(
        description='Prepare SIDD test dataset',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--data-root', help='dataset root')
    parser.add_argument('--out-dir', help='output directory of dataset')
    args = parser.parse_args()
    return args


if __name__ == '__main__':
    args = parse_args()

    # export images
    export_images(args)



# config for model
stable_diffusion_v15_url = 'runwayml/stable-diffusion-v1-5'
models_path = '/home/AnimateDiff/models/'
randomness = dict(
    seed=[
        10917152860782582783, 6399018107401806238, 15875751942533906793,
        6653196880059936551
    ],
    diff_rank_seed=True)

diffusion_scheduler = dict(
    type='DDIMScheduler',
    beta_end=0.012,
    beta_schedule='linear',
    beta_start=0.00085,
    num_train_timesteps=1000,
    prediction_type='epsilon',
    set_alpha_to_one=True,
    clip_sample=False,
    thresholding=False,
    steps_offset=1)

model = dict(
    type='AnimateDiff',
    vae=dict(
        type='AutoencoderKL',
        from_pretrained=stable_diffusion_v15_url,
        subfolder='vae'),
    unet=dict(
        type='UNet3DConditionMotionModel',
        unet_use_cross_frame_attention=False,
        unet_use_temporal_attention=False,
        use_motion_module=True,
        motion_module_resolutions=[1, 2, 4, 8],
        motion_module_mid_block=False,
        motion_module_decoder_only=False,
        motion_module_type='Vanilla',
        motion_module_kwargs=dict(
            num_attention_heads=8,
            num_transformer_block=1,
            attention_block_types=['Temporal_Self', 'Temporal_Self'],
            temporal_position_encoding=True,
            temporal_position_encoding_max_len=24,
            temporal_attention_dim_div=1),
        subfolder='unet',
        from_pretrained=stable_diffusion_v15_url),
    text_encoder=dict(
        type='ClipWrapper',
        clip_type='huggingface',
        pretrained_model_name_or_path=stable_diffusion_v15_url,
        subfolder='text_encoder'),
    tokenizer=stable_diffusion_v15_url,
    scheduler=diffusion_scheduler,
    test_scheduler=diffusion_scheduler,
    data_preprocessor=dict(type='DataPreprocessor'),
    motion_module_cfg=dict(path=models_path + 'Motion_Module/mm_sd_v14.ckpt'),
    dream_booth_lora_cfg=dict(
        type='ToonYou',
        path=models_path + 'DreamBooth_LoRA/lyriel_v16.safetensors',
        steps=25,
        guidance_scale=7.5))



# define GAN model

d_reg_interval = 16
g_reg_interval = 4

g_reg_ratio = g_reg_interval / (g_reg_interval + 1)
d_reg_ratio = d_reg_interval / (d_reg_interval + 1)

loss_config = dict(
    r1_loss_weight=10. / 2. * d_reg_interval,
    r1_interval=d_reg_interval,
    norm_mode='HWC',
    g_reg_interval=g_reg_interval,
    g_reg_weight=2. * g_reg_interval,
    pl_batch_shrink=2)

model = dict(
    type='StyleGAN2',
    data_preprocessor=dict(type='DataPreprocessor'),
    generator=dict(
        type='StyleGANv2Generator',
        out_size=None,  # Need to be set.
        style_channels=512,
    ),
    discriminator=dict(
        type='StyleGAN2Discriminator',
        in_size=None,  # Need to be set.
    ),
    ema_config=dict(type='ExponentialMovingAverage'),
    loss_config=loss_config)



# Copyright (c) OpenMMLab. All rights reserved.
from .deblurganv2 import DeblurGanV2
from .deblurganv2_discriminator import DeblurGanV2Discriminator
from .deblurganv2_generator import DeblurGanV2Generator

__all__ = ['DeblurGanV2', 'DeblurGanV2Generator', 'DeblurGanV2Discriminator']



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Callable, List, Optional, Sequence, Union

import numpy as np
import torch
from mmengine import is_list_of
from torch import Tensor


def noise_sample_fn(noise: Union[Tensor, Callable, None] = None,
                    *,
                    num_batches: int = 1,
                    noise_size: Union[int, Sequence[int], None] = None,
                    device: Optional[str] = None) -> Tensor:
    """Sample noise with respect to the given `num_batches`, `noise_size` and
    `device`.

    Args:
        noise (torch.Tensor | callable | None): You can directly give a
            batch of noise through a ``torch.Tensor`` or offer a callable
            function to sample a batch of noise data. Otherwise, the ``None``
            indicates to use the default noise sampler. Defaults to None.
        num_batches (int, optional): The number of batch size. Defaults to 1.
        noise_size (Union[int, Sequence[int], None], optional): The size of
            random noise. Defaults to None.
        device (Optional[str], optional): The target device of the random
            noise. Defaults to None.

    Returns:
        Tensor: Sampled random noise.
    """
    if isinstance(noise, torch.Tensor):
        noise_batch = noise
        # unsqueeze if need
        if noise_batch.ndim == 1:
            noise_batch = noise_batch[None, ...]
    else:
        # generate noise automatically, prepare and check `num_batches` and
        # `noise_size`
        # num_batches = 1 if num_batches is None else num_batches
        assert num_batches > 0, (
            'If you want to generate noise automatically, \'num_batches\' '
            'must larger than 0.')
        assert noise_size is not None, (
            'If you want to generate noise automatically, \'noise_size\' '
            'must not be None.')
        noise_size = [noise_size] if isinstance(noise_size, int) \
            else noise_size

        if callable(noise):
            # receive a noise generator and sample noise.
            noise_generator = noise
            noise_batch = noise_generator((num_batches, *noise_size))
        else:
            # otherwise, we will adopt default noise sampler.
            assert num_batches > 0
            noise_batch = torch.randn((num_batches, *noise_size))

    # Check the shape if `noise_size` is passed. Ignore `num_batches` here
    # because `num_batches` has default value.
    if noise_size is not None:
        if isinstance(noise_size, int):
            noise_size = [noise_size]
        assert list(noise_batch.shape[1:]) == list(noise_size), (
            'Size of the input noise is inconsistency with \'noise_size\'\'. '
            f'Receive \'{noise_batch.shape[1:]}\' and \'{noise_size}\' '
            'respectively.')

    if device is not None:
        return noise_batch.to(device)
    return noise_batch


def label_sample_fn(label: Union[Tensor, Callable, List[int], None] = None,
                    *,
                    num_batches: int = 1,
                    num_classes: Optional[int] = None,
                    device: Optional[str] = None) -> Union[Tensor, None]:
    """Sample random label with respect to `num_batches`, `num_classes` and
    `device`.

    Args:
        label (Union[Tensor, Callable, List[int], None], optional): You can
            directly give a batch of label through a ``torch.Tensor`` or offer
            a callable function to sample a batch of label data. Otherwise, the
            ``None`` indicates to use the default label sampler.
            Defaults to None.
        num_batches (int, optional): The number of batch size. Defaults to 1.
        num_classes (Optional[int], optional): The number of classes. Defaults
            to None.
        device (Optional[str], optional): The target device of the label.
            Defaults to None.

    Returns:
        Union[Tensor, None]: Sampled random label.
    """
    # label is not passed and do not have `num_classes` to sample label
    if (num_classes is None or num_classes <= 0) and (label is None):
        return None

    if isinstance(label, Tensor):
        label_batch = label
    elif isinstance(label, np.ndarray):
        label_batch = torch.from_numpy(label).long()
    elif isinstance(label, list):
        if is_list_of(label, (int, np.ndarray)):
            label = [torch.LongTensor([lab]).squeeze() for lab in label]
        else:
            assert is_list_of(label, torch.Tensor), (
                'Only support \'int\', \'np.ndarray\' and \'torch.Tensor\' '
                'type for list input. But receives '
                f'\'{[type(lab) for lab in label]}\'')
            label = [lab.squeeze() for lab in label]
        label_batch = torch.stack(label, dim=0)
    else:
        # generate label_batch manually, prepare and check `num_batches`
        assert num_batches > 0, (
            'If you want to generate label automatically, \'num_batches\' '
            'must larger than 0.')

        if callable(label):
            # receive a noise generator and sample noise.
            label_generator = label
            label_batch = label_generator(num_batches)
        else:
            # otherwise, we will adopt default label sampler.
            label_batch = torch.randint(0, num_classes, (num_batches, ))

    # Check whether is LongTensor (torch.int64) and shape like [bz, ]
    assert label_batch.dtype == torch.int64, (
        'Input label cannot convert to torch.LongTensor (torch.int64). Please '
        'check your input.')
    assert label_batch.ndim == 1, (
        'Input label must shape like \'[num_batches, ]\', but shape like '
        f'({label_batch.shape})')

    # Check the label if `num_classes` is passed. Ignore `num_batches` because
    # `num_batches` has default value.
    if num_classes is not None:
        invalid_index = torch.logical_or(label_batch < 0,
                                         label_batch >= num_classes)
        assert not (invalid_index).any(), (
            f'Labels in label_batch must be in range [0, num_classes - 1] '
            f'([0, {num_classes}-1]). But found {label_batch[invalid_index]} '
            'are out of range.')

    if device is not None:
        return label_batch.to(device)
    return label_batch



# Copyright (c) OpenMMLab. All rights reserved.
from enum import Enum

import torch
import torch.nn as nn
from diffusers.models.attention import BasicTransformerBlock
from torch import Tensor

AttentionStatus = Enum('ATTENTION_STATUS', 'READ WRITE DISABLE')


def torch_dfs(model: torch.nn.Module):
    result = [model]
    for child in model.children():
        result += torch_dfs(child)
    return result


class AttentionInjection(nn.Module):
    """Wrapper for stable diffusion unet.

    Args:
        module (nn.Module): The module to be wrapped.
    """

    def __init__(self, module: nn.Module, injection_weight=5):
        super().__init__()
        self.attention_status = AttentionStatus.READ
        self.style_cfgs = []
        self.unet = module

        attn_inject = self

        def transformer_forward_replacement(
            self,
            hidden_states,
            attention_mask=None,
            encoder_hidden_states=None,
            encoder_attention_mask=None,
            timestep=None,
            cross_attention_kwargs=None,
            class_labels=None,
        ):
            if self.use_ada_layer_norm:
                norm_hidden_states = self.norm1(hidden_states, timestep)
            elif self.use_ada_layer_norm_zero:
                norm_hidden_states, gate_msa, shift_mlp, scale_mlp, gate_mlp = self.norm1(  # noqa
                    hidden_states,
                    timestep,
                    class_labels,
                    hidden_dtype=hidden_states.dtype)
            else:
                norm_hidden_states = self.norm1(hidden_states)

            attn_output = None
            self_attention_context = norm_hidden_states
            if attn_inject.attention_status == AttentionStatus.WRITE:
                self.bank.append(self_attention_context.detach().clone())
            if attn_inject.attention_status == AttentionStatus.READ:
                if len(self.bank) > 0:
                    self.bank = self.bank * injection_weight
                    attn_output = self.attn1(
                        norm_hidden_states,
                        encoder_hidden_states=torch.cat(
                            [self_attention_context] + self.bank, dim=1))
                    # attn_output = self.attn1(
                    #     norm_hidden_states,
                    #     encoder_hidden_states=self.bank[0])
                self.bank = []
            if attn_output is None:
                attn_output = self.attn1(norm_hidden_states)

            if self.use_ada_layer_norm_zero:
                attn_output = gate_msa.unsqueeze(1) * attn_output
            hidden_states = attn_output + hidden_states

            cross_attention_kwargs = cross_attention_kwargs if \
                cross_attention_kwargs is not None else {}
            if self.attn2 is not None:
                norm_hidden_states = (
                    self.norm2(hidden_states, timestep)
                    if self.use_ada_layer_norm else self.norm2(hidden_states))

                # 2. Cross-Attention
                attn_output = self.attn2(
                    norm_hidden_states,
                    encoder_hidden_states=encoder_hidden_states,
                    attention_mask=encoder_attention_mask,
                    **cross_attention_kwargs,
                )
                hidden_states = attn_output + hidden_states

            # 3. Feed-forward
            norm_hidden_states = self.norm3(hidden_states)

            if self.use_ada_layer_norm_zero:
                norm_hidden_states = norm_hidden_states * \
                    (1 + scale_mlp[:, None]) + shift_mlp[:, None]

            ff_output = self.ff(norm_hidden_states)

            if self.use_ada_layer_norm_zero:
                ff_output = gate_mlp.unsqueeze(1) * ff_output

            hidden_states = ff_output + hidden_states

            return hidden_states

        all_modules = torch_dfs(self.unet)

        attn_modules = [
            module for module in all_modules
            if isinstance(module, BasicTransformerBlock)
        ]
        for i, module in enumerate(attn_modules):
            if getattr(module, '_original_inner_forward', None) is None:
                module._original_inner_forward = module.forward
            module.forward = transformer_forward_replacement.__get__(
                module, BasicTransformerBlock)
            module.bank = []

    def forward(self,
                x: Tensor,
                t,
                encoder_hidden_states=None,
                down_block_additional_residuals=None,
                mid_block_additional_residual=None,
                ref_x=None) -> Tensor:
        """Forward and add LoRA mapping.

        Args:
            x (Tensor): The input tensor.

        Returns:
            Tensor: The output tensor.
        """
        if ref_x is not None:
            self.attention_status = AttentionStatus.WRITE
            self.unet(
                ref_x,
                t,
                encoder_hidden_states=encoder_hidden_states,
                down_block_additional_residuals=  # noqa
                down_block_additional_residuals,
                mid_block_additional_residual=mid_block_additional_residual)
        self.attention_status = AttentionStatus.READ
        output = self.unet(
            x,
            t,
            encoder_hidden_states=encoder_hidden_states,
            down_block_additional_residuals=  # noqa
            down_block_additional_residuals,
            mid_block_additional_residual=mid_block_additional_residual)

        return output



# Base config for CelebA-HQ dataset

# dataset settings
dataset_type = 'BasicImageDataset'
data_root = 'data/CelebA-HQ'

train_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    sampler=dict(type='InfiniteSampler', shuffle=True),
    dataset=dict(
        type=dataset_type,
        data_root=data_root,
        data_prefix=dict(gt=''),
        ann_file='train_celeba_img_list.txt',
        test_mode=False,
    ))

val_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type=dataset_type,
        data_root=data_root,
        data_prefix=dict(gt=''),
        ann_file='val_celeba_img_list.txt',
        test_mode=True,
    ))

test_dataloader = val_dataloader

val_evaluator = [
    dict(type='MAE', mask_key='mask', scaling=100),
    # By default, compute with pixel value from 0-1
    # scale=2 to align with 1.0
    # scale=100 seems to align with readme
    dict(type='PSNR'),
    dict(type='SSIM'),
]

test_evaluator = val_evaluator



_base_ = '../_base_/gen_default_runtime.py'

# config for model
stable_diffusion_v15_url = 'runwayml/stable-diffusion-v1-5'
controlnet_canny_url = 'lllyasviel/sd-controlnet-canny'

model = dict(
    type='ControlStableDiffusion',
    vae=dict(
        type='AutoencoderKL',
        from_pretrained=stable_diffusion_v15_url,
        subfolder='vae'),
    unet=dict(
        type='UNet2DConditionModel',
        subfolder='unet',
        from_pretrained=stable_diffusion_v15_url),
    text_encoder=dict(
        type='ClipWrapper',
        clip_type='huggingface',
        pretrained_model_name_or_path=stable_diffusion_v15_url,
        subfolder='text_encoder'),
    tokenizer=stable_diffusion_v15_url,
    controlnet=dict(
        type='ControlNetModel',
        # from_pretrained=controlnet_canny_rul
        from_config=controlnet_canny_url  # train from scratch
    ),
    scheduler=dict(
        type='DDPMScheduler',
        from_pretrained=stable_diffusion_v15_url,
        subfolder='scheduler'),
    test_scheduler=dict(
        type='DDIMScheduler',
        from_pretrained=stable_diffusion_v15_url,
        subfolder='scheduler'),
    data_preprocessor=dict(type='DataPreprocessor'),
    init_cfg=dict(type='init_from_unet'))

# config for training
train_cfg = dict(max_iters=10000)
optim_wrapper = dict(controlnet=dict(optimizer=dict(type='AdamW', lr=1e-5)))

# Config for data loader
pipeline = [
    dict(type='LoadImageFromFile', key='source', channel_order='rgb'),
    dict(type='LoadImageFromFile', key='target', channel_order='rgb'),
    dict(
        type='PackInputs',
        keys=['source', 'target'],
        data_keys='prompt',
        meta_keys=[
            'source_channel_order', 'source_color_type',
            'target_channel_order', 'target_color_type'
        ])
]
dataset = dict(
    type='ControlNetDataset',
    data_root='./data/fill50k',
    ann_file='prompt.json',
    pipeline=pipeline)
train_dataloader = dict(
    dataset=dataset,
    num_workers=16,
    sampler=dict(type='InfiniteSampler', shuffle=True),
    persistent_workers=True,
    batch_size=4)
val_cfg = val_evaluator = val_dataloader = None
test_cfg = test_evaluator = test_dataloader = None

# hooks
custom_hooks = [
    dict(
        type='VisualizationHook',
        interval=300,
        fixed_input=True,
        # visualize train dataset
        vis_kwargs_list=dict(type='Data', name='fake_img'),
        n_samples=4,
        n_row=2)
]



# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy
from typing import List, Optional

import mmengine
import torch
import torch.nn as nn
import torch.nn.functional as F
from mmengine import MessageHub
from mmengine.model import BaseModel, is_model_wrapper
from mmengine.optim import OptimWrapperDict
from mmengine.runner.checkpoint import _load_checkpoint_with_prefix
from tqdm import tqdm

from mmagic.registry import DIFFUSION_SCHEDULERS, MODELS
from mmagic.structures import DataSample
from mmagic.utils.typing import ForwardInputs, SampleList


def classifier_grad(classifier, x, t, y=None, classifier_scale=1.0):
    """compute classification gradient to x."""
    assert y is not None
    with torch.enable_grad():
        x_in = x.detach().requires_grad_(True)
        timesteps = torch.ones_like(y) * t
        logits = classifier(x_in, timesteps)
        log_probs = F.log_softmax(logits, dim=-1)
        selected = log_probs[range(len(logits)), y.view(-1)]
        return torch.autograd.grad(selected.sum(), x_in)[0] * classifier_scale


@MODELS.register_module('ADM')
@MODELS.register_module('GuidedDiffusion')
@MODELS.register_module()
class AblatedDiffusionModel(BaseModel):
    """Guided diffusion Model.

    Args:
        data_preprocessor (dict, optional): The pre-process config of
            :class:`BaseDataPreprocessor`.
        unet (ModelType): Config of denoising Unet.
        diffusion_scheduler (ModelType): Config of diffusion_scheduler
            scheduler.
        use_fp16 (bool): Whether to use fp16 for unet model. Defaults to False.
        classifier (ModelType): Config of classifier. Defaults to None.
        pretrained_cfgs (dict): Path Config for pretrained weights. Usually
            this is a dict contains module name and the corresponding ckpt
            path.Defaults to None.
    """

    def __init__(self,
                 data_preprocessor,
                 unet,
                 diffusion_scheduler,
                 use_fp16=False,
                 classifier=None,
                 classifier_scale=1.0,
                 rgb2bgr=False,
                 pretrained_cfgs=None):

        super().__init__(data_preprocessor=data_preprocessor)
        self.unet = MODELS.build(unet)
        self.diffusion_scheduler = DIFFUSION_SCHEDULERS.build(
            diffusion_scheduler)
        if classifier:
            self.classifier = MODELS.build(classifier)
        else:
            self.classifier = None
        self.classifier_scale = classifier_scale

        if pretrained_cfgs:
            self.load_pretrained_models(pretrained_cfgs)
        if use_fp16:
            mmengine.print_log('Convert unet modules to floatpoint16')
            self.unet.convert_to_fp16()
        self.rgb2bgr = rgb2bgr

    def load_pretrained_models(self, pretrained_cfgs):
        """_summary_

        Args:
            pretrained_cfgs (_type_): _description_
        """
        for key, ckpt_cfg in pretrained_cfgs.items():
            prefix = ckpt_cfg.get('prefix', '')
            map_location = ckpt_cfg.get('map_location', 'cpu')
            strict = ckpt_cfg.get('strict', True)
            ckpt_path = ckpt_cfg.get('ckpt_path')
            if prefix == '':
                state_dict = torch.load(ckpt_path, map_location=map_location)
            else:
                state_dict = _load_checkpoint_with_prefix(
                    prefix, ckpt_path, map_location)
            getattr(self, key).load_state_dict(state_dict, strict=strict)
            mmengine.print_log(f'Load pretrained {key} from {ckpt_path}')

    @property
    def device(self):
        """Get current device of the model.

        Returns:
            torch.device: The current device of the model.
        """
        return next(self.parameters()).device

    @torch.no_grad()
    def infer(self,
              scheduler_kwargs=None,
              init_image=None,
              batch_size=1,
              num_inference_steps=1000,
              labels=None,
              classifier_scale=0.0,
              show_progress=False):
        """_summary_

        Args:
            init_image (_type_, optional): _description_. Defaults to None.
            batch_size (int, optional): _description_. Defaults to 1.
            num_inference_steps (int, optional): _description_.
                Defaults to 1000.
            labels (_type_, optional): _description_. Defaults to None.
            show_progress (bool, optional): _description_. Defaults to False.

        Returns:
            _type_: _description_
        """
        if scheduler_kwargs is not None:
            mmengine.print_log('Switch to infer diffusion scheduler!',
                               'current')
            infer_scheduler = DIFFUSION_SCHEDULERS.build(scheduler_kwargs)
        else:
            infer_scheduler = self.diffusion_scheduler

        # Sample gaussian noise to begin loop
        if init_image is None:
            image = torch.randn(
                (batch_size, self.get_module(self.unet, 'in_channels'),
                 self.get_module(self.unet, 'image_size'),
                 self.get_module(self.unet, 'image_size')))
        else:
            image = init_image
        image = image.to(self.device)

        if isinstance(labels, int):
            labels = torch.tensor(labels).repeat(batch_size)
        elif labels is None:
            labels = torch.randint(
                low=0,
                high=self.get_module(self.unet, 'num_classes'),
                size=(batch_size, ),
                device=self.device)
        labels = labels.to(self.device)

        # set step values
        if num_inference_steps > 0:
            infer_scheduler.set_timesteps(num_inference_steps)

        timesteps = infer_scheduler.timesteps

        if show_progress and mmengine.dist.is_main_process():
            timesteps = tqdm(timesteps)
        for t in timesteps:
            # 1. predicted model_output
            model_output = self.unet(image, t, label=labels)['sample']

            # 2. compute previous image: x_t -> x_t-1
            if classifier_scale > 0 and self.classifier is not None:
                cond_fn = classifier_grad
                cond_kwargs = dict(
                    y=labels,
                    classifier=self.classifier,
                    classifier_scale=classifier_scale)
            else:
                cond_fn = None
                cond_kwargs = {}
            diffusion_scheduler_output = infer_scheduler.step(
                model_output,
                t,
                image,
                cond_fn=cond_fn,
                cond_kwargs=cond_kwargs)
            image = diffusion_scheduler_output['prev_sample']

        if self.rgb2bgr:
            image = image[:, [2, 1, 0], ...]

        return {'samples': image}

    def forward(self,
                inputs: ForwardInputs,
                data_samples: Optional[list] = None,
                mode: Optional[str] = None) -> List[DataSample]:
        """_summary_

        Args:
            inputs (ForwardInputs): _description_
            data_samples (Optional[list], optional): _description_.
                Defaults to None.
            mode (Optional[str], optional): _description_. Defaults to None.

        Returns:
            List[DataSample]: _description_
        """
        init_image = inputs.get('init_image', None)
        batch_size = inputs.get('num_batches', 1)
        sample_kwargs = inputs.get('sample_kwargs', dict())

        labels = sample_kwargs.get('labels', None)
        num_inference_steps = sample_kwargs.get(
            'num_inference_steps',
            self.diffusion_scheduler.num_train_timesteps)
        show_progress = sample_kwargs.get('show_progress', False)
        classifier_scale = sample_kwargs.get('classifier_scale',
                                             self.classifier_scale)

        outputs = self.infer(
            init_image=init_image,
            batch_size=batch_size,
            num_inference_steps=num_inference_steps,
            show_progress=show_progress,
            classifier_scale=classifier_scale)

        batch_sample_list = []
        for idx in range(batch_size):
            gen_sample = DataSample()
            if data_samples:
                gen_sample.update(data_samples[idx])
            if isinstance(outputs, dict):
                gen_sample.fake_img = outputs['samples'][idx]
                gen_sample.set_gt_label(labels[idx])

            # Append input condition (noise and sample_kwargs) to
            # batch_sample_list
            if init_image is not None:
                gen_sample.noise = init_image[idx]
            gen_sample.sample_kwargs = deepcopy(sample_kwargs)
            batch_sample_list.append(gen_sample)
        return batch_sample_list

    def val_step(self, data: dict) -> SampleList:
        """Gets the generated image of given data.

        Calls ``self.data_preprocessor(data)`` and
        ``self(inputs, data_sample, mode=None)`` in order. Return the
        generated results which will be passed to evaluator.

        Args:
            data (dict): Data sampled from metric specific
                sampler. More details in `Metrics` and `Evaluator`.

        Returns:
            SampleList: Generated image or image dict.
        """
        data = self.data_preprocessor(data)
        outputs = self(**data)
        return outputs

    def test_step(self, data: dict) -> SampleList:
        """Gets the generated image of given data. Same as :meth:`val_step`.

        Args:
            data (dict): Data sampled from metric specific
                sampler. More details in `Metrics` and `Evaluator`.

        Returns:
            List[DataSample]: Generated image or image dict.
        """
        data = self.data_preprocessor(data)
        outputs = self(**data)
        return outputs

    def train_step(self, data: dict, optim_wrapper: OptimWrapperDict):
        """_summary_

        Args:
            data (dict): _description_
            optim_wrapper (OptimWrapperDict): _description_

        Returns:
            _type_: _description_
        """
        message_hub = MessageHub.get_current_instance()
        curr_iter = message_hub.get_info('iter')

        # sampling x0 and timestep
        data = self.data_preprocessor(data)
        real_imgs = data['inputs']
        timestep = self.diffusion_scheduler.sample_timestep()

        # calculating loss
        loss_dict = self.diffusion_scheduler.training_loss(
            self.unet, real_imgs, timestep)
        loss, log_vars = self._parse_losses(loss_dict)
        optim_wrapper['denoising'].update_params(loss)

        # update EMA
        if self.with_ema_denoising and (curr_iter + 1) >= self.ema_start:
            self.denoising_ema.update_parameters(
                self.denoising_ema.
                module if is_model_wrapper(self.denoising) else self.denoising)
            # if not update buffer, copy buffer from orig model
            if not self.denoising_ema.update_buffers:
                self.denoising_ema.sync_buffers(
                    self.denoising.module
                    if is_model_wrapper(self.denoising) else self.denoising)
        elif self.with_ema_denoising:
            # before ema, copy weights from orig
            self.denoising_ema.sync_parameters(
                self.denoising.
                module if is_model_wrapper(self.denoising) else self.denoising)

        return log_vars

    def get_module(self, model: nn.Module, module_name: str) -> nn.Module:
        """Get an inner module from model.

        Since we will wrapper DDP for some model, we have to judge whether the
        module can be indexed directly.

        Args:
            model (nn.Module): This model may wrapped with DDP or not.
            module_name (str): The name of specific module.

        Return:
            nn.Module: Returned sub module.
        """
        module = model.module if hasattr(model, 'module') else model
        return getattr(module, module_name)



"""Config for the `config-f` setting in StyleGAN2."""

_base_ = ['./stylegan2_c2_8xb4-800kiters_ffhq-256x256.py']

model = dict(loss_config=dict(r1_use_apex_amp=False, g_reg_use_apex_amp=False))

train_cfg = dict(max_iters=800002)

# remain to be refactored
apex_amp = dict(mode='gan', init_args=dict(opt_level='O1', num_losses=2))
resume_from = None



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional, Sequence

from mmengine import MessageHub
from mmengine.hooks import ParamSchedulerHook
from mmengine.runner import Runner

from mmagic.registry import HOOKS

DATA_BATCH = Optional[Sequence[dict]]


@HOOKS.register_module()
class ReduceLRSchedulerHook(ParamSchedulerHook):
    """A hook to update learning rate.

    Args:
        val_metric (str): The metric of validation. If val_metric is not None,
            we check val_metric to reduce learning. Default: None.
        by_epoch (bool): Whether to update by epoch. Default: True.
        interval (int): The interval of iterations to update. Default: 1.
    """

    def __init__(self,
                 val_metric: str = None,
                 by_epoch=True,
                 interval=1) -> None:
        super().__init__()

        self.message_hub = MessageHub.get_instance('reduce_lr')

        self.val_metric = val_metric
        self.by_epoch = by_epoch
        self.interval = interval
        self.sum_value = 0
        self.count = 0

    def _calculate_average_value(self):
        value = self.sum_value / self.count
        self.sum_value = 0
        self.count = 0
        self.message_hub.update_scalar('value', value)

    def after_train_epoch(self, runner: Runner):
        """Call step function for each scheduler after each train epoch.

        Args:
            runner (Runner): The runner of the training process.
        """
        if not self.by_epoch:
            return

        # If val_metric is not None, we check val_metric to reduce learning
        if self.val_metric is not None:
            return

        if self.every_n_epochs(runner, self.interval):
            self._calculate_average_value()
            super().after_train_epoch(runner=runner)

    def after_train_iter(self,
                         runner: Runner,
                         batch_idx: int,
                         data_batch: DATA_BATCH = None,
                         outputs: Optional[dict] = None) -> None:
        """Call step function for each scheduler after each iteration.

        Args:
            runner (Runner): The runner of the training process.
            batch_idx (int): The index of the current batch in the train loop.
            data_batch (Sequence[dict], optional): Data from dataloader.
                In order to keep this interface consistent with other hooks,
                we keep ``data_batch`` here. Defaults to None.
            outputs (dict, optional): Outputs from model.
                In order to keep this interface consistent with other hooks, we
                keep ``data_batch`` here. Defaults to None.
        """

        # If val_metric is not None, we check val_metric to reduce learning
        if self.val_metric is not None:
            return

        current = runner.message_hub.get_scalar('train/loss').current()
        self.sum_value += current * len(data_batch)
        self.count += len(data_batch)

        if self.by_epoch:
            return

        if self.every_n_train_iters(runner, self.interval):
            self._calculate_average_value()
            super().after_train_iter(
                runner=runner,
                batch_idx=batch_idx,
                data_batch=data_batch,
                outputs=outputs)

    def after_val_epoch(self,
                        runner,
                        metrics: Optional[Dict[str, float]] = None):
        """Call step function for each scheduler after each validation epoch.

        Args:
            runner (Runner): The runner of the training process.
            metrics (dict, optional): The metrics of validation. Default: None.
        """
        # If val_metric is None, we check training loss to reduce learning
        # rate.
        if self.val_metric is None:
            return

        if self.val_metric not in metrics:
            raise KeyError(f'{self.val_metric} is not found in metrics')

        self.sum_value += metrics[self.val_metric]
        self.count += 1

        if not self.by_epoch or self.every_n_epochs(runner, self.interval):
            # if self.by_epoch is False,
            # call val after several iter
            # and update LR in each ``after_val_epoch``
            self._calculate_average_value()

            def step(param_schedulers):
                assert isinstance(param_schedulers, list)
                for scheduler in param_schedulers:
                    scheduler.step()

            if isinstance(runner.param_schedulers, list):
                step(runner.param_schedulers)
            elif isinstance(runner.param_schedulers, dict):
                for param_schedulers in runner.param_schedulers.values():
                    step(param_schedulers)
            else:
                raise TypeError(
                    'runner.param_schedulers should be list of ParamScheduler '
                    'or a dict containing list of ParamScheduler, '
                    f'but got {runner.param_schedulers}')



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
import torch.nn.functional as F

from mmagic.registry import MODELS
from .deblurganv2_util import MobileNetV2, get_norm_layer, inceptionresnetv2

backbone_list = ['FPNInception', 'FPNMobileNet', 'FPNInceptionSimple']


class FPNHead(nn.Module):
    """Head for FPNInception,FPNInceptionSimple and FPNMobilenet."""

    def __init__(self, num_in, num_mid, num_out):
        super().__init__()

        self.block0 = nn.Conv2d(
            num_in, num_mid, kernel_size=3, padding=1, bias=False)
        self.block1 = nn.Conv2d(
            num_mid, num_out, kernel_size=3, padding=1, bias=False)

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        x = nn.functional.relu(self.block0(x), inplace=True)
        x = nn.functional.relu(self.block1(x), inplace=True)
        return x


class FPN_inception(nn.Module):

    def __init__(self, norm_layer, num_filter=256, pretrained='imagenet'):
        """Creates an `FPN` instance for feature extraction.

        Args:
          num_filter: the number of filters in each output pyramid level
          pretrained: use ImageNet pre-trained backbone feature extractor
        """

        super().__init__()
        self.inception = inceptionresnetv2(
            num_classes=1000, pretrained=pretrained)

        self.enc0 = self.inception.conv2d_1a
        self.enc1 = nn.Sequential(
            self.inception.conv2d_2a,
            self.inception.conv2d_2b,
            self.inception.maxpool_3a,
        )  # 64
        self.enc2 = nn.Sequential(
            self.inception.conv2d_3b,
            self.inception.conv2d_4a,
            self.inception.maxpool_5a,
        )  # 192
        self.enc3 = nn.Sequential(
            self.inception.mixed_5b,
            self.inception.repeat,
            self.inception.mixed_6a,
        )  # 1088
        self.enc4 = nn.Sequential(
            self.inception.repeat_1,
            self.inception.mixed_7a,
        )  # 2080
        self.td1 = nn.Sequential(
            nn.Conv2d(num_filter, num_filter, kernel_size=3, padding=1),
            norm_layer(num_filter), nn.ReLU(inplace=True))
        self.td2 = nn.Sequential(
            nn.Conv2d(num_filter, num_filter, kernel_size=3, padding=1),
            norm_layer(num_filter), nn.ReLU(inplace=True))
        self.td3 = nn.Sequential(
            nn.Conv2d(num_filter, num_filter, kernel_size=3, padding=1),
            norm_layer(num_filter), nn.ReLU(inplace=True))
        self.pad = nn.ReflectionPad2d(1)
        self.lateral4 = nn.Conv2d(2080, num_filter, kernel_size=1, bias=False)
        self.lateral3 = nn.Conv2d(1088, num_filter, kernel_size=1, bias=False)
        self.lateral2 = nn.Conv2d(192, num_filter, kernel_size=1, bias=False)
        self.lateral1 = nn.Conv2d(64, num_filter, kernel_size=1, bias=False)
        self.lateral0 = nn.Conv2d(
            32, num_filter // 2, kernel_size=1, bias=False)

        for param in self.inception.parameters():
            param.requires_grad = False

    def unfreeze(self):
        """Unfreeze params."""
        for param in self.inception.parameters():
            param.requires_grad = True

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        # Bottom-up pathway, from ResNet
        enc0 = self.enc0(x)

        enc1 = self.enc1(enc0)  # 256

        enc2 = self.enc2(enc1)  # 512

        enc3 = self.enc3(enc2)  # 1024

        enc4 = self.enc4(enc3)  # 2048

        # Lateral connections

        lateral4 = self.pad(self.lateral4(enc4))
        lateral3 = self.pad(self.lateral3(enc3))
        lateral2 = self.lateral2(enc2)
        lateral1 = self.pad(self.lateral1(enc1))
        lateral0 = self.lateral0(enc0)

        # Top-down pathway
        pad = (1, 2, 1, 2)  # pad last dim by 1 on each side
        pad1 = (0, 1, 0, 1)
        map4 = lateral4
        map3 = self.td1(
            lateral3 +
            nn.functional.upsample(map4, scale_factor=2, mode='nearest'))
        map2 = self.td2(
            F.pad(lateral2, pad, 'reflect') +
            nn.functional.upsample(map3, scale_factor=2, mode='nearest'))
        map1 = self.td3(
            lateral1 +
            nn.functional.upsample(map2, scale_factor=2, mode='nearest'))
        return F.pad(lateral0, pad1, 'reflect'), map1, map2, map3, map4


class FPNInception(nn.Module):
    """Feature Pyramid Network (FPN) with four feature maps of resolutions 1/4,
    1/8, 1/16, 1/32 and `num_filter` filters for all feature maps."""

    def __init__(self,
                 norm_layer,
                 output_ch=3,
                 num_filter=128,
                 num_filter_fpn=256):
        super().__init__()

        norm_layer = get_norm_layer(norm_type=norm_layer)
        self.fpn = FPN_inception(
            num_filter=num_filter_fpn, norm_layer=norm_layer)

        # The segmentation heads on top of the FPN

        self.head1 = FPNHead(num_filter_fpn, num_filter, num_filter)
        self.head2 = FPNHead(num_filter_fpn, num_filter, num_filter)
        self.head3 = FPNHead(num_filter_fpn, num_filter, num_filter)
        self.head4 = FPNHead(num_filter_fpn, num_filter, num_filter)

        self.smooth = nn.Sequential(
            nn.Conv2d(4 * num_filter, num_filter, kernel_size=3, padding=1),
            norm_layer(num_filter),
            nn.ReLU(),
        )

        self.smooth2 = nn.Sequential(
            nn.Conv2d(num_filter, num_filter // 2, kernel_size=3, padding=1),
            norm_layer(num_filter // 2),
            nn.ReLU(),
        )

        self.final = nn.Conv2d(
            num_filter // 2, output_ch, kernel_size=3, padding=1)

    def unfreeze(self):
        """Unfreeze params."""
        self.fpn.unfreeze()

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        map0, map1, map2, map3, map4 = self.fpn(x)

        map4 = nn.functional.upsample(
            self.head4(map4), scale_factor=8, mode='nearest')
        map3 = nn.functional.upsample(
            self.head3(map3), scale_factor=4, mode='nearest')
        map2 = nn.functional.upsample(
            self.head2(map2), scale_factor=2, mode='nearest')
        map1 = nn.functional.upsample(
            self.head1(map1), scale_factor=1, mode='nearest')

        smoothed = self.smooth(torch.cat([map4, map3, map2, map1], dim=1))
        smoothed = nn.functional.upsample(
            smoothed, scale_factor=2, mode='nearest')
        smoothed = self.smooth2(smoothed + map0)
        smoothed = nn.functional.upsample(
            smoothed, scale_factor=2, mode='nearest')

        final = self.final(smoothed)
        res = torch.tanh(final) + x

        return torch.clamp(res, min=-1, max=1)


class FPN_inceptionsimple(nn.Module):

    def __init__(self, norm_layer, num_filters=256):
        """Creates an `FPN` instance for feature extraction.

        Args:
          num_filters: the number of filters in each output pyramid level
          pretrained: use ImageNet pre-trained backbone feature extractor
        """

        super().__init__()
        self.inception = inceptionresnetv2(
            num_classes=1000, pretrained='imagenet')

        self.enc0 = self.inception.conv2d_1a
        self.enc1 = nn.Sequential(
            self.inception.conv2d_2a,
            self.inception.conv2d_2b,
            self.inception.maxpool_3a,
        )  # 64
        self.enc2 = nn.Sequential(
            self.inception.conv2d_3b,
            self.inception.conv2d_4a,
            self.inception.maxpool_5a,
        )  # 192
        self.enc3 = nn.Sequential(
            self.inception.mixed_5b,
            self.inception.repeat,
            self.inception.mixed_6a,
        )  # 1088
        self.enc4 = nn.Sequential(
            self.inception.repeat_1,
            self.inception.mixed_7a,
        )  # 2080

        self.pad = nn.ReflectionPad2d(1)
        self.lateral4 = nn.Conv2d(2080, num_filters, kernel_size=1, bias=False)
        self.lateral3 = nn.Conv2d(1088, num_filters, kernel_size=1, bias=False)
        self.lateral2 = nn.Conv2d(192, num_filters, kernel_size=1, bias=False)
        self.lateral1 = nn.Conv2d(64, num_filters, kernel_size=1, bias=False)
        self.lateral0 = nn.Conv2d(
            32, num_filters // 2, kernel_size=1, bias=False)

        for param in self.inception.parameters():
            param.requires_grad = False

    def unfreeze(self):
        """Unfreeze params."""
        for param in self.inception.parameters():
            param.requires_grad = True

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        # Bottom-up pathway, from ResNet
        enc0 = self.enc0(x)

        enc1 = self.enc1(enc0)  # 256

        enc2 = self.enc2(enc1)  # 512

        enc3 = self.enc3(enc2)  # 1024

        enc4 = self.enc4(enc3)  # 2048

        # Lateral connections

        lateral4 = self.pad(self.lateral4(enc4))
        lateral3 = self.pad(self.lateral3(enc3))
        lateral2 = self.lateral2(enc2)
        lateral1 = self.pad(self.lateral1(enc1))
        lateral0 = self.lateral0(enc0)

        # Top-down pathway
        pad = (1, 2, 1, 2)  # pad last dim by 1 on each side
        pad1 = (0, 1, 0, 1)
        map4 = lateral4
        map3 = lateral3 + nn.functional.upsample(
            map4, scale_factor=2, mode='nearest')
        map2 = F.pad(lateral2, pad, 'reflect') + nn.functional.upsample(
            map3, scale_factor=2, mode='nearest')
        map1 = lateral1 + nn.functional.upsample(
            map2, scale_factor=2, mode='nearest')
        return F.pad(lateral0, pad1, 'reflect'), map1, map2, map3, map4


class FPNInceptionSimple(nn.Module):
    """Feature Pyramid Network (FPN) with four feature maps of resolutions 1/4,
    1/8, 1/16, 1/32 and `num_filter` filters for all feature maps."""

    def __init__(self,
                 norm_layer,
                 output_ch=3,
                 num_filter=128,
                 num_filter_fpn=256):
        super().__init__()

        norm_layer = get_norm_layer(norm_type=norm_layer)
        self.fpn = FPN_inceptionsimple(
            num_filter=num_filter_fpn, norm_layer=norm_layer)

        # The segmentation heads on top of the FPN

        self.head1 = FPNHead(num_filter_fpn, num_filter, num_filter)
        self.head2 = FPNHead(num_filter_fpn, num_filter, num_filter)
        self.head3 = FPNHead(num_filter_fpn, num_filter, num_filter)
        self.head4 = FPNHead(num_filter_fpn, num_filter, num_filter)

        self.smooth = nn.Sequential(
            nn.Conv2d(4 * num_filter, num_filter, kernel_size=3, padding=1),
            norm_layer(num_filter),
            nn.ReLU(),
        )

        self.smooth2 = nn.Sequential(
            nn.Conv2d(num_filter, num_filter // 2, kernel_size=3, padding=1),
            norm_layer(num_filter // 2),
            nn.ReLU(),
        )

        self.final = nn.Conv2d(
            num_filter // 2, output_ch, kernel_size=3, padding=1)

    def unfreeze(self):
        """unfreeze the fpn network."""
        self.fpn.unfreeze()

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        map0, map1, map2, map3, map4 = self.fpn(x)

        map4 = nn.functional.upsample(
            self.head4(map4), scale_factor=8, mode='nearest')
        map3 = nn.functional.upsample(
            self.head3(map3), scale_factor=4, mode='nearest')
        map2 = nn.functional.upsample(
            self.head2(map2), scale_factor=2, mode='nearest')
        map1 = nn.functional.upsample(
            self.head1(map1), scale_factor=1, mode='nearest')

        smoothed = self.smooth(torch.cat([map4, map3, map2, map1], dim=1))
        smoothed = nn.functional.upsample(
            smoothed, scale_factor=2, mode='nearest')
        smoothed = self.smooth2(smoothed + map0)
        smoothed = nn.functional.upsample(
            smoothed, scale_factor=2, mode='nearest')

        final = self.final(smoothed)
        res = torch.tanh(final) + x

        return torch.clamp(res, min=-1, max=1)


class FPN_mobilenet(nn.Module):

    def __init__(self, norm_layer, num_filters=128, pretrained=None):
        """Creates an `FPN` instance for feature extraction.

        Args:
          num_filters: the number of filters in each output pyramid level
          pretrained: use ImageNet pre-trained backbone feature extractor
        """

        super().__init__()
        net = MobileNetV2(n_class=1000)

        if pretrained:
            # Load weights into the project directory
            if torch.cuda.is_available():
                state_dict = torch.load(
                    pretrained)  # add map_location='cpu' if no gpu
            else:
                state_dict = torch.load(pretrained, map_location='cpu')
            net.load_state_dict(state_dict)
        self.features = net.features

        self.enc0 = nn.Sequential(*self.features[0:2])
        self.enc1 = nn.Sequential(*self.features[2:4])
        self.enc2 = nn.Sequential(*self.features[4:7])
        self.enc3 = nn.Sequential(*self.features[7:11])
        self.enc4 = nn.Sequential(*self.features[11:16])

        self.td1 = nn.Sequential(
            nn.Conv2d(num_filters, num_filters, kernel_size=3, padding=1),
            norm_layer(num_filters), nn.ReLU(inplace=True))
        self.td2 = nn.Sequential(
            nn.Conv2d(num_filters, num_filters, kernel_size=3, padding=1),
            norm_layer(num_filters), nn.ReLU(inplace=True))
        self.td3 = nn.Sequential(
            nn.Conv2d(num_filters, num_filters, kernel_size=3, padding=1),
            norm_layer(num_filters), nn.ReLU(inplace=True))

        self.lateral4 = nn.Conv2d(160, num_filters, kernel_size=1, bias=False)
        self.lateral3 = nn.Conv2d(64, num_filters, kernel_size=1, bias=False)
        self.lateral2 = nn.Conv2d(32, num_filters, kernel_size=1, bias=False)
        self.lateral1 = nn.Conv2d(24, num_filters, kernel_size=1, bias=False)
        self.lateral0 = nn.Conv2d(
            16, num_filters // 2, kernel_size=1, bias=False)

        for param in self.features.parameters():
            param.requires_grad = False

    def unfreeze(self):
        """Unfreeze params."""
        for param in self.features.parameters():
            param.requires_grad = True

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        # Bottom-up pathway, from ResNet
        enc0 = self.enc0(x)

        enc1 = self.enc1(enc0)  # 256

        enc2 = self.enc2(enc1)  # 512

        enc3 = self.enc3(enc2)  # 1024

        enc4 = self.enc4(enc3)  # 2048

        # Lateral connections

        lateral4 = self.lateral4(enc4)
        lateral3 = self.lateral3(enc3)
        lateral2 = self.lateral2(enc2)
        lateral1 = self.lateral1(enc1)
        lateral0 = self.lateral0(enc0)

        # Top-down pathway
        map4 = lateral4
        map3 = self.td1(
            lateral3 +
            nn.functional.upsample(map4, scale_factor=2, mode='nearest'))
        map2 = self.td2(
            lateral2 +
            nn.functional.upsample(map3, scale_factor=2, mode='nearest'))
        map1 = self.td3(
            lateral1 +
            nn.functional.upsample(map2, scale_factor=2, mode='nearest'))
        return lateral0, map1, map2, map3, map4


class FPNMobileNet(nn.Module):

    def __init__(self,
                 norm_layer,
                 output_ch=3,
                 num_filter=64,
                 num_filter_fpn=128,
                 pretrained=None):
        super().__init__()

        # Feature Pyramid Network (FPN) with four feature maps of resolutions
        # 1/4, 1/8, 1/16, 1/32 and `num_filters` filters for all feature maps.
        norm_layer = get_norm_layer(norm_type=norm_layer)
        self.fpn = FPN_mobilenet(
            num_filters=num_filter_fpn,
            norm_layer=norm_layer,
            pretrained=pretrained)

        # The segmentation heads on top of the FPN

        self.head1 = FPNHead(num_filter_fpn, num_filter, num_filter)
        self.head2 = FPNHead(num_filter_fpn, num_filter, num_filter)
        self.head3 = FPNHead(num_filter_fpn, num_filter, num_filter)
        self.head4 = FPNHead(num_filter_fpn, num_filter, num_filter)

        self.smooth = nn.Sequential(
            nn.Conv2d(4 * num_filter, num_filter, kernel_size=3, padding=1),
            norm_layer(num_filter),
            nn.ReLU(),
        )

        self.smooth2 = nn.Sequential(
            nn.Conv2d(num_filter, num_filter // 2, kernel_size=3, padding=1),
            norm_layer(num_filter // 2),
            nn.ReLU(),
        )

        self.final = nn.Conv2d(
            num_filter // 2, output_ch, kernel_size=3, padding=1)

    def unfreeze(self):
        """unfreeze the fpn network."""
        self.fpn.unfreeze()

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        map0, map1, map2, map3, map4 = self.fpn(x)

        map4 = nn.functional.upsample(
            self.head4(map4), scale_factor=8, mode='nearest')
        map3 = nn.functional.upsample(
            self.head3(map3), scale_factor=4, mode='nearest')
        map2 = nn.functional.upsample(
            self.head2(map2), scale_factor=2, mode='nearest')
        map1 = nn.functional.upsample(
            self.head1(map1), scale_factor=1, mode='nearest')

        smoothed = self.smooth(torch.cat([map4, map3, map2, map1], dim=1))
        smoothed = nn.functional.upsample(
            smoothed, scale_factor=2, mode='nearest')
        smoothed = self.smooth2(smoothed + map0)
        smoothed = nn.functional.upsample(
            smoothed, scale_factor=2, mode='nearest')

        final = self.final(smoothed)
        res = torch.tanh(final) + x

        return torch.clamp(res, min=-1, max=1)


@MODELS.register_module()
class DeblurGanV2Generator:
    """Defines the generator for DeblurGanv2 with the specified arguments..

    Args:
        model (Str): Type of the generator  model
    """

    def __new__(cls, backbone, *args, **kwargs):
        if backbone == 'FPNInception':
            return FPNInception(*args, **kwargs)
        elif backbone == 'FPNMobileNet':
            return FPNMobileNet(*args, **kwargs)
        elif backbone == 'FPNInceptionSimple':
            return FPNInceptionSimple(*args, **kwargs)
        else:
            raise Exception('Generator model {} not found, '
                            'Please use the following models: '
                            '{}'.format(backbone, backbone_list))



_base_ = [
    '../_base_/default_runtime.py',
    '../_base_/datasets/deraining_test_config.py'
]

experiment_name = 'restormer_official_rain13k'
work_dir = f'./work_dirs/{experiment_name}'
save_dir = './work_dirs/'

# model settings
model = dict(
    type='BaseEditModel',
    generator=dict(
        type='Restormer',
        inp_channels=3,
        out_channels=3,
        dim=48,
        num_blocks=[4, 6, 6, 8],
        num_refinement_blocks=4,
        heads=[1, 2, 4, 8],
        ffn_expansion_factor=2.66,
        bias=False,
        LayerNorm_type='WithBias',
        dual_pixel_task=False),
    pixel_loss=dict(type='L1Loss', loss_weight=1.0, reduction='mean'),
    data_preprocessor=dict(
        type='DataPreprocessor',
        mean=[0., 0., 0.],
        std=[255., 255., 255.],
    ))
